import os
import sys
import copy
import shutil
import typing
import logging
import pathlib

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import rez.config
import rez.packages
import rez.package_maker
import rez.resolved_context
import rez.vendor.version.version

import rez_pip.pip
import rez_pip.utils

_LOG = logging.getLogger(__name__)


def createPackage(
    dist: importlib_metadata.Distribution,
    isPure: bool,
    pythonVersion: rez.vendor.version.version.Version,
    nameCasings: typing.List[str],
    installedWheelsDir: str,
    wheelURL: str,
    prefix: typing.Optional[str] = None,
    release: bool = False,
) -> None:
    _LOG.info(f"Creating rez package for {dist.name}")
    name = rez_pip.utils.pythontDistributionNameToRez(dist.name)
    version = rez_pip.utils.pythonDistributionVersionToRez(dist.version)

    requirements = rez_pip.utils.getRezRequirements(dist, pythonVersion, isPure, [])

    requires = requirements.requires
    variant_requires = requirements.variant_requires
    metadata = requirements.metadata

    if prefix:
        packagesPath = prefix
    else:
        packagesPath = (
            rez.config.config.release_packages_path
            if release
            else rez.config.config.local_packages_path
        )

    def make_root(variant: rez.packages.Variant, path: str) -> None:
        """Using distlib to iterate over all installed files of the current
        distribution to copy files to the target directory of the rez package
        variant
        """
        formattedRequirements = ", ".join(str(req) for req in variant.variant_requires)

        _LOG.info(
            rf"Installing {variant.qualified_package_name} \[{formattedRequirements}]"
        )
        if not dist.files:
            raise RuntimeError(
                f"{dist.name} package has no files registered! Something is wrong maybe?"
            )

        wheelsDirAbsolute = pathlib.Path(installedWheelsDir).resolve()
        for src in dist.files:
            srcAbsolute = src.locate().resolve()
            dest = os.path.join(path, srcAbsolute.relative_to(wheelsDirAbsolute))
            if not os.path.exists(os.path.dirname(dest)):
                os.makedirs(os.path.dirname(dest))

            _LOG.debug(f"Copying {str(srcAbsolute)!r} to {str(dest)!r}")
            shutil.copyfile(srcAbsolute, dest)
            shutil.copystat(srcAbsolute, dest)

    with rez.package_maker.make_package(
        name, packagesPath, make_root=make_root, skip_existing=True, warn_on_skip=False
    ) as pkg:
        pkg.version = version

        # requirements and variants
        if requires:
            pkg.requires = requires

        if variant_requires:
            pkg.variants = [variant_requires]

        # commands
        commands = ["env.PYTHONPATH.append('{root}/python')"]

        console_scripts = [
            ep.name for ep in dist.entry_points if ep.group == "console_scripts"
        ]
        if console_scripts:
            pkg.tools = console_scripts
            # TODO: Don't hardcode scripts here.
            commands.append("env.PATH.append('{root}/scripts')")

        pkg.commands = "\n".join(commands)

        # Make the package use hashed variants. This is required because we
        # can't control what ends up in its variants, and that can easily
        # include problematic chars (>, +, ! etc).
        # TODO: #672 (shortlinks for variants)
        pkg.hashed_variants = True

        pkg.pip = {
            "name": dist.name,
            "version": dist.version,
            "is_pure_python": metadata["is_pure_python"],
            "wheel_url": wheelURL,
        }

        # Take all the metadata that can be converted and put it
        # in the rez package definition.
        convertedMetadata, remainingMetadata = _convertMetadata(dist)
        for key, values in convertedMetadata.items():
            setattr(pkg, key, values)

        pkg.pip["metadata"] = remainingMetadata

    _LOG.info(
        f"[bold]Created {len(pkg.installed_variants)} variants and skipped {len(pkg.skipped_variants)}"
    )


def _convertMetadata(
    dist: importlib_metadata.Distribution,
) -> typing.Tuple[typing.Dict[str, typing.Any], typing.Dict[str, typing.Any]]:
    metadata = {}
    originalMetadata = copy.deepcopy(dist.metadata.json)
    del originalMetadata["metadata_version"]
    del originalMetadata["name"]
    del originalMetadata["version"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#summary
    if dist.metadata["Summary"]:
        metadata["summary"] = dist.metadata["Summary"]
        del originalMetadata["summary"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#description
    if dist.metadata["Description"]:
        metadata["description"] = dist.metadata["Description"]
        del originalMetadata["description"]

    authors = []

    # https://packaging.python.org/en/latest/specifications/core-metadata/#author
    author = dist.metadata["Author"]
    if author:
        authors.append(author)
        del originalMetadata["author"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#author-email
    authorEmail = dist.metadata["Author-email"]
    if authorEmail:
        authors.extend([email.strip() for email in authorEmail.split(",")])
        del originalMetadata["author_email"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#maintainer
    maintainer = dist.metadata["Maintainer"]
    if maintainer:
        authors.append(maintainer)
        del originalMetadata["maintainer"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#maintainer-email
    maintainerEmail = dist.metadata["Maintainer-email"]
    if maintainerEmail:
        authors.extend([email.strip() for email in maintainerEmail.split(",")])
        del originalMetadata["maintainer_email"]

    if authors:
        metadata["authors"] = authors

    # https://packaging.python.org/en/latest/specifications/core-metadata/#license
    # Prefer the License field and fallback to classifiers if one is present.
    if dist.metadata["License"]:
        metadata["license"] = dist.metadata["License"]
        del originalMetadata["license"]
    else:
        licenseClassifiers = [
            item.split("::")[-1].strip()
            for item in dist.metadata.get_all("classifier", [])
            if item.startswith("License ::")
        ]

        # I don't know what to do in this case, so just skip license if more than one
        # classifier is found.
        if len(licenseClassifiers) == 1:
            metadata["license"] = licenseClassifiers[0]

    helpLinks = []

    # https://packaging.python.org/en/latest/specifications/core-metadata/#home-page
    if dist.metadata["Home-page"]:
        helpLinks.append(["Home-page", dist.metadata["Home-page"]])
        del originalMetadata["home_page"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#project-url-multiple-use
    if dist.metadata["Project-URL"]:
        urls = [
            url.strip()
            for value in dist.metadata.get_all("Project-URL")
            for url in value.split(",")
        ]
        helpLinks.extend([list(entry) for entry in zip(urls[::2], urls[1::2])])
        del originalMetadata["project_url"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#download-url
    if dist.metadata["Download-URL"]:
        helpLinks.append(["Download-URL", dist.metadata["Download-URL"]])
        del originalMetadata["download_url"]

    if helpLinks:
        metadata["help"] = helpLinks

    return metadata, originalMetadata


def getPythonExecutables(
    range_: typing.Optional[str], packageFamily: str = "python"
) -> typing.Dict[str, pathlib.Path]:
    """
    Get the available python executable from rez packages.

    :param range_: version specifier
    :param packageFamily: Name of the rez package family for the python package. This allows ot support PyPy, etc.
    :returns: Dict where the keys are the python versions and values are abolute paths to executables.
    """
    packages = sorted(
        rez.packages.iter_packages(
            packageFamily, range_=range_ if range_ != "latest" else None
        ),
        key=lambda x: x.version,
    )

    if range_ == "latest":
        packages = [packages[-1]]

    pythons: typing.Dict[str, pathlib.Path] = {}
    for package in packages:
        resolvedContext = rez.resolved_context.ResolvedContext(
            [f"{package.name}=={package.version}"]
        )

        # Make sure that system PATH doens't interfere with the "which" method.
        resolvedContext.append_sys_path = False

        for trimmedVersion in map(package.version.trim, [2, 1, 0]):
            path = resolvedContext.which(f"python{trimmedVersion}", parent_environ={})
            if path:
                pythons[str(package.version)] = pathlib.Path(path)
                break
        else:
            _LOG.warning(
                f"Failed to find a Python executable in the {package.qualified_name!r} rez package"
            )

    return pythons



import rich
import rich.console


class RezPipError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message

    def __rich_console__(
        self, console: rich.console.Console, options: rich.console.ConsoleOptions
    ) -> rich.console.RenderResult:
        yield console.render_str(
            f"[bold]{self.__class__.__module__}.{self.__class__.__name__}[/]: {self.message}"
        )


class PipError(RezPipError):
    pass



import os
import sys
import glob
import hashlib
import pathlib
import platform
import subprocess
import collections

import rez.packages
import rez.resolved_context
import rez.vendor.version.version


def getPythonRezPackageExecutablePath(version: str, repo: str):
    package = rez.packages.get_package("python", version, paths=[repo])
    assert package

    ctx = rez.resolved_context.ResolvedContext(
        [package.qualified_name], package_paths=[repo]
    )

    executable = f"python{str(package.version).split('.')[0]}"
    if platform.system() == "Windows":
        executable = "python.exe"

    return ctx.which(executable), ctx


def buildPackage(name: str, outputDir: str) -> pathlib.Path:
    sourcePath = os.path.join(os.path.dirname(__file__), "data", "src_packages", name)

    subprocess.run(
        [
            sys.executable,
            "-m",
            "build",
            "-w",
            ".",
            "--outdir",
            outputDir,
            "--no-isolation",
        ],
        cwd=sourcePath,
        check=True,
    )

    return pathlib.Path(glob.glob(os.path.join(outputDir, f"{name}*.whl"))[0])


class PyPIIndex:
    def __init__(self, path: pathlib.Path):
        self._path = path
        self._cache = collections.defaultdict(dict)

    @property
    def path(self) -> pathlib.Path:
        return self._path

    def getWheel(self, name: str) -> pathlib.Path:
        return next(self.path.joinpath(name).glob("*.whl"))

    def getWheelHash(self, name: str, typ="sha256") -> str:
        path = self.getWheel(name)

        cachedValue = self._cache[typ].get(path)
        if cachedValue:
            return cachedValue

        buf = bytearray(2**18)  # Reusable buffer to reduce allocations.
        view = memoryview(buf)

        digestobj = hashlib.new(typ)
        with open(path, "rb") as fd:
            while True:
                size = fd.readinto(buf)
                if size == 0:
                    break  # EOF
                digestobj.update(view[:size])

        digest = digestobj.hexdigest()
        self._cache[typ][path] = digest
        return digest



import os
import sys
import typing
import asyncio
import hashlib
import logging

import rich
import aiohttp
import rich.progress

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import rez_pip.pip

_LOG = logging.getLogger(__name__)
_lock = asyncio.Lock()


def downloadPackages(
    packages: typing.List[rez_pip.pip.PackageInfo], dest: str
) -> typing.List[str]:
    return asyncio.run(_downloadPackages(packages, dest))


async def _downloadPackages(
    packages: typing.List[rez_pip.pip.PackageInfo], dest: str
) -> typing.List[str]:
    items: typing.List[
        typing.Coroutine[typing.Any, typing.Any, typing.Optional[str]]
    ] = []
    wheels = []

    async with aiohttp.ClientSession() as session:
        with rich.progress.Progress(
            "[progress.description]{task.description}",
            "[progress.percentage]{task.percentage:>3.0f}%",
            rich.progress.BarColumn(),
            rich.progress.DownloadColumn(),
            rich.progress.TransferSpeedColumn(),
            transient=True,
            console=rich.get_console(),
        ) as progress:
            tasks: typing.Dict[str, rich.progress.TaskID] = {}

            # Create all the downlod tasks first
            for package in packages:
                tasks[package.name] = progress.add_task(package.name)

            # Then create the "total" progress bar. This ensures that total is at the bottom.
            mainTask = progress.add_task(f"[bold]Total (0/{len(packages)})", total=0)

            for package in packages:
                items.append(
                    _download(
                        package, dest, session, progress, tasks[package.name], mainTask
                    )
                )

            wheels = await asyncio.gather(*items)

    if not all(wheels):
        raise RuntimeError("Some wheels failed to be downloaded")

    return typing.cast(typing.List[str], wheels)


def getSHA256(path: str) -> str:
    buf = bytearray(2**18)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)

    digestobj = hashlib.new("sha256")

    with open(path, "rb") as fd:
        while True:
            size = fd.readinto(buf)
            if size == 0:
                break  # EOF
            digestobj.update(view[:size])

    return digestobj.hexdigest()


async def _download(
    package: rez_pip.pip.PackageInfo,
    target: str,
    session: aiohttp.ClientSession,
    progress: rich.progress.Progress,
    taskID: rich.progress.TaskID,
    mainTaskID: rich.progress.TaskID,
) -> typing.Optional[str]:
    wheelName: str = os.path.basename(package.download_info.url)
    wheelPath = os.path.join(target, wheelName)

    # TODO: Handle case where sha256 doesn't exist. We should also support the other supported
    # hash types.
    if (
        os.path.exists(wheelPath)
        and getSHA256(wheelPath) == package.download_info.archive_info.hashes["sha256"]
    ):
        _LOG.info(f"{wheelName} found in cache at {wheelPath!r}. Skipping download.")
    else:
        _LOG.debug(
            f"Downloading {package.name}-{package.version} from {package.download_info.url}"
        )

        async with session.get(
            package.download_info.url,
            headers={
                "Content-Type": "application/octet-stream",
                "User-Agent": f"rez-pip/{importlib_metadata.version('rez-pip')}",
            },
        ) as response:
            size = int(response.headers.get("content-length", 0))
            progress.update(taskID, total=size)

            async with _lock:
                mainTask = [task for task in progress.tasks if task.id == mainTaskID][0]

                progress.update(
                    mainTaskID,
                    total=typing.cast(int, mainTask.total) + size,
                )

            if response.status != 200:
                _LOG.error(
                    f"failed to download {package.download_info.url}: {response.status} - {response.reason}, {response.request_info}"
                )
                return None

            with open(wheelPath, "wb") as fd:
                async for chunk, asd in response.content.iter_chunks():
                    if not chunk:
                        break
                    progress.update(taskID, advance=len(chunk))
                    progress.update(mainTaskID, advance=len(chunk))
                    fd.write(chunk)

            _LOG.info(
                f"Downloaded {package.name}-{package.version} to {wheelPath!r} ({os.stat(wheelPath).st_size} bytes)"
            )

    progress.update(taskID, visible=False)

    total = len(progress.tasks) - 1
    async with _lock:
        completedItems = [task for task in progress.tasks if not task.visible]
        progress.update(
            mainTaskID, description=f"[bold]Total ({len(completedItems)}/{total})"
        )

    return wheelPath



import os
import time
import typing
import pathlib
import zipfile
import platform
import textwrap
import subprocess
import http.client
import urllib.request

import pytest
import rez.config
import rez.packages
import rez.package_bind
import rez.package_maker

from . import utils

DATA_ROOT_DIR = os.path.join(os.path.dirname(__file__), "data")
DOWNLOAD_DIR = os.path.join(DATA_ROOT_DIR, "_tmp_download")

phaseReportKey = pytest.StashKey[typing.Dict[str, pytest.CollectReport]]()


# Token from https://docs.pytest.org/en/latest/example/simple.html#making-test-result-information-available-in-fixtures
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item: pytest.Item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # store test results for each phase of a call, which can
    # be "setup", "call", "teardown"
    item.stash.setdefault(phaseReportKey, {})[rep.when] = rep


@pytest.fixture(scope="session")
def index(tmpdir_factory: pytest.TempdirFactory) -> utils.PyPIIndex:
    """Build PyPI Index and return the path"""

    srcPackages = os.path.join(DATA_ROOT_DIR, "src_packages")

    indexPath = tmpdir_factory.mktemp("pypi_index").dirpath()

    for pkg in os.listdir(srcPackages):
        dest = indexPath.mkdir(pkg)
        utils.buildPackage(pkg, os.fspath(dest))

    return utils.PyPIIndex(pathlib.Path(indexPath.strpath))


@pytest.fixture(scope="function")
def pypi(
    printer_session: typing.Callable[[str], None],
    index: utils.PyPIIndex,
    request: pytest.FixtureRequest,
) -> typing.Generator[str, None, None]:
    """Start a PyPI instance and return the URL to talk to it."""
    port = 45678
    host = "localhost"

    proc = subprocess.Popen(
        [
            "pypi-server",
            "run",
            os.fspath(index.path),
            f"--port={port}",
            f"--host={host}",
            "--disable-fallback",
            "--log-stream=stdout",
            "--hash-algo=sha256",  # Defaults to md5
            "-v",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    url = f"http://{host}:{port}"

    retries = 50
    while retries > 0:
        conn = http.client.HTTPConnection(f"{host}:{port}")
        try:
            conn.request("HEAD", "/health")
            response = conn.getresponse()
            if response is not None:
                print(response.read())
                yield url
                break
        except ConnectionRefusedError:
            time.sleep(0.1)
            retries -= 1

    proc.terminate()
    proc.wait()

    output = textwrap.indent(proc.stdout.read(), "\t\t")

    if proc.returncode and proc.returncode != -15 and platform.system() != "Windows":
        pytest.fail(
            f"pypi-server returned a non zero error code ({proc.returncode}):\n\n{proc.stdout.read()}"
        )

    if (
        request.node.stash[phaseReportKey].get("call")
        and request.node.stash[phaseReportKey]["call"].failed
    ):
        printer_session(f"pypi-server output:\n{output}")

    if not retries:
        raise RuntimeError("Failed to start pypi-server")


# Use function scope to make sure each test starts with a fresh default config
@pytest.fixture(scope="function", autouse=True)
def hardenRezConfig(tmp_path_factory: pytest.TempPathFactory):
    """Make sure that rez doesn't use any of the user configs"""

    # We can't just put None, so override with a dummy directory to make sure
    # that we don't pick up packages from anywhere else.
    tmpRepoPath = tmp_path_factory.mktemp("tmp_fake_rez_repo_do_not_use")

    # Do not load user configs
    defaultConfig = rez.config._create_locked_config(
        {
            "packages_path": [os.fspath(tmpRepoPath)],
            "release_packages_path": os.fspath(tmpRepoPath),
            "local_packages_path": os.fspath(tmpRepoPath),
        }
    )
    with rez.config._replace_config(defaultConfig):
        yield


@pytest.fixture(scope="session")
def rezRepo() -> typing.Generator[str, None, None]:
    path = os.path.join(DATA_ROOT_DIR, "rez_repo")

    rez.package_bind.bind_package("platform", path=path, no_deps=True, quiet=True)
    rez.package_bind.bind_package("arch", path=path, no_deps=True, quiet=True)
    rez.package_bind.bind_package("os", path=path, no_deps=True, quiet=True)

    originalConfig = rez.config.config.copy()

    rez.config.config.packages_path = [path]
    rez.config.config.release_packages_path = path
    yield path

    rez.config.config.packages_path = originalConfig.packages_path
    rez.config.config.release_packages_path = originalConfig.release_packages_path


def downloadPythonVersion(
    version: str, printer_session: typing.Callable[[str], None]
) -> str:
    """
    Download a Python interpreter for a given version

    :param version: Python version to download.
    :param printer_session: Pytest session printer.
    :returns: Path to downloaded artifact.
    """
    try:
        os.makedirs(DOWNLOAD_DIR)
    except FileExistsError:
        pass

    url = f"https://globalcdn.nuget.org/packages/python.{version}.nupkg"

    path = os.path.join(
        DOWNLOAD_DIR, f"python-{version}-{platform.system().lower()}.nupkg"
    )

    if os.path.exists(path):
        printer_session(f"Skipping {url} because {path!r} already exists")
        return path

    printer_session(f"Downloading {url} to {path!r}")
    with urllib.request.urlopen(url) as archive:
        with open(path, "wb") as targetFile:
            targetFile.write(archive.read())

    return path


@pytest.fixture(
    scope="session",
    params=[
        pytest.param(
            # Nuget doesn't have 3.7.16
            "3.7.9" if platform.system() == "Windows" else "3.7.16",
            marks=pytest.mark.py37,
        ),
        pytest.param(
            # Nuget doesn't have 3.9.16
            "3.9.13" if platform.system() == "Windows" else "3.9.16",
            marks=pytest.mark.py39,
        ),
        pytest.param("3.11.3", marks=pytest.mark.py311),
    ],
)
def pythonRezPackage(
    request: pytest.FixtureRequest,
    rezRepo: str,
    printer_session: typing.Callable[[str], None],
) -> str:
    """
    Create a Python rez package and return Python version number.
    """
    version = typing.cast(str, request.param)

    def make_root(variant: rez.packages.Variant, path: str) -> None:
        """Using distlib to iterate over all installed files of the current
        distribution to copy files to the target directory of the rez package
        variant
        """
        printer_session(f"Creating rez package for Python {version} in {rezRepo!r}")

        dest = os.path.join(path, "python")

        if platform.system() == "Windows":
            archivePath = downloadPythonVersion(version, printer_session)
            with zipfile.ZipFile(archivePath) as archive:
                printer_session(f"Extracting {archivePath!r} to {dest!r}")
                archive.extractall(path=dest)
        else:
            import conda.cli.python_api

            printer_session(
                f"Installing Python {version} by creating a conda environment at {dest!r}"
            )
            conda.cli.python_api.run_command(
                conda.cli.python_api.Commands.CREATE,
                "--prefix",
                dest,
                f"conda-forge:python={version}",
                "pip",
            )

    with rez.package_maker.make_package(
        "python",
        rezRepo,
        make_root=make_root,
        skip_existing=True,
        warn_on_skip=False,
    ) as pkg:
        pkg.version = version

        commands = [
            "env.PATH.prepend('{root}/python/bin')",
        ]
        if platform.system() == "Windows":
            commands = [
                "env.PATH.prepend('{root}/python/tools')",
                "env.PATH.prepend('{root}/python/tools/DLLs')",
            ]

        pkg.commands = "\n".join(commands)

    if pkg.skipped_variants:
        printer_session(
            f"Python {version} rez package already exists at {pkg.skipped_variants[0].uri}"
        )

    return version



import os
import sys
import stat
import typing
import pathlib
import platform
import unittest.mock

import pytest
import rez.config
import rez.packages
import rez.package_repository

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import rez_pip.rez


def test_convertMetadata_nothing_to_convert(monkeypatch: pytest.MonkeyPatch):
    dist = importlib_metadata.Distribution()
    monkeypatch.setattr(
        dist,
        "read_text",
        lambda x: "Metadata-Version: 2.0\nName: package_a\nVersion: 1.0.0",
    )

    converted, remaining = rez_pip.rez._convertMetadata(dist)
    assert converted == {}
    assert remaining == {}

    # Assert that the original metadata is intact
    assert dist.metadata["Metadata-Version"]


@pytest.mark.parametrize(
    "metadataText,expectedConverted,expectedRemaining",
    [
        ["Summary: This is a summary", {"summary": "This is a summary"}, {}],
        [
            "Description: This is a description",
            {"description": "This is a description"},
            {},
        ],
        ["Author: Joe", {"authors": ["Joe"]}, {}],
        [
            "Author-email: Joe <joe@example.com>",
            {"authors": ["Joe <joe@example.com>"]},
            {},
        ],
        [
            "Author-email: Joe <joe@example.com>,Joe2 <joe2@example.com> , Joe3 asd <joe3@example.com> ",
            {
                "authors": [
                    "Joe <joe@example.com>",
                    "Joe2 <joe2@example.com>",
                    "Joe3 asd <joe3@example.com>",
                ]
            },
            {},
        ],
        ["Maintainer: Example", {"authors": ["Example"]}, {}],
        [
            "Maintainer-email: Example <example@example.com>",
            {"authors": ["Example <example@example.com>"]},
            {},
        ],
        [
            "Maintainer-email: Joe <joe@example.com>,Joe2 <joe2@example.com> , Joe3 asd <joe3@example.com> ",
            {
                "authors": [
                    "Joe <joe@example.com>",
                    "Joe2 <joe2@example.com>",
                    "Joe3 asd <joe3@example.com>",
                ]
            },
            {},
        ],
        [
            "Author: Jo1\nAuthor-email: asd@example.com\nMaintainer: <something@example.com>\nMaintainer-email: Joe <joe@example.com>",
            {
                "authors": [
                    "Jo1",
                    "asd@example.com",
                    "<something@example.com>",
                    "Joe <joe@example.com>",
                ]
            },
            {},
        ],
        # License from License field
        ["License: Apache-2.0", {"license": "Apache-2.0"}, {}],
        [
            "Classifier: asdasd :: something",
            {},
            {"classifier": ["asdasd :: something"]},
        ],
        # License from classifier
        [
            "Classifier: License :: Some license",
            {"license": "Some license"},
            {"classifier": ["License :: Some license"]},
        ],
        # When reading license from classifiers, skip if more than one license found
        [
            "Classifier: License :: Some license1\nClassifier: License :: Some license2",
            {},
            {"classifier": ["License :: Some license1", "License :: Some license2"]},
        ],
        [
            "Home-page: https://example.com",
            {"help": [["Home-page", "https://example.com"]]},
            {},
        ],
        [
            "Project-URL: Documentation, https://example.com/docs",
            {"help": [["Documentation", "https://example.com/docs"]]},
            {},
        ],
        [
            "Project-URL: Documentation, https://example.com/docs\nProject-URL: Other, https://example.com/other",
            {
                "help": [
                    ["Documentation", "https://example.com/docs"],
                    ["Other", "https://example.com/other"],
                ]
            },
            {},
        ],
        [
            "Download-URL: https://example.com/download",
            {"help": [["Download-URL", "https://example.com/download"]]},
            {},
        ],
        # Test additive links
        [
            "Home-page: https://example.com/home\nProject-URL: Documentation, https://example.com/docs\nProject-URL: Other, https://example.com/other\nDownload-URL: https://example.com/download",
            {
                "help": [
                    ["Home-page", "https://example.com/home"],
                    ["Documentation", "https://example.com/docs"],
                    ["Other", "https://example.com/other"],
                    ["Download-URL", "https://example.com/download"],
                ],
            },
            {},
        ],
        # Test remaining values
        [
            "Author: me\nTotally-random: some value here",
            {"authors": ["me"]},
            {"totally_random": "some value here"},
        ],
    ],
)
def test_convertMetadata(
    metadataText: str,
    expectedConverted,
    expectedRemaining,
    monkeypatch: pytest.MonkeyPatch,
):
    dist = importlib_metadata.Distribution()
    monkeypatch.setattr(
        dist,
        "read_text",
        lambda x: f"Metadata-Version: 2.0\nName: package_a\nVersion: 1.0.0\n{metadataText}",
    )

    converted, remaining = rez_pip.rez._convertMetadata(dist)
    assert converted == expectedConverted
    assert remaining == expectedRemaining

    # Assert that the original metadata is intact
    assert dist.metadata["Metadata-Version"]


@pytest.mark.parametrize(
    "availableVersions,range_,executables,expectedExecutables",
    [
        pytest.param(
            ["1.0.0"],
            "latest",
            ["python1.0"],
            {"1.0.0": pathlib.Path("/path/python1.0")},
            id="latest",
        ),
        pytest.param(
            ["1.0.0"],
            None,
            ["python1.0"],
            {"1.0.0": pathlib.Path("/path/python1.0")},
            id="all-single-version",
        ),
        pytest.param(
            ["1.0.0"],
            None,
            ["", "python1"],
            {"1.0.0": pathlib.Path("/path/python1")},
            id="all-single-version-python1",
        ),
        pytest.param(
            ["1.0.0"],
            None,
            ["", "", "python"],
            {"1.0.0": pathlib.Path("/path/python")},
            id="all-single-version-python-no-version",
        ),
        pytest.param(
            ["1.0.0"],
            None,
            ["", "", ""],
            {},
            id="all-single-version-no-exe",
        ),
        pytest.param(
            ["1.0.0", "2.0.0"],
            None,
            ["python1.0", "python2.0"],
            {
                "1.0.0": pathlib.Path("/path/python1.0"),
                "2.0.0": pathlib.Path("/path/python2.0"),
            },
            id="all-multi-version",
        ),
        pytest.param(
            ["1.0.0", "2.0.0"],
            None,
            ["", "", "", "python2.0"],
            {"2.0.0": pathlib.Path("/path/python2.0")},
            id="all-multi-version-no-exe-1.0.0",
        ),
        pytest.param(
            ["1.0.0", "2.0.0"],
            "2+",
            ["python2.0"],
            {"2.0.0": pathlib.Path("/path/python2.0")},
            id="with-range-2plus",
        ),
        pytest.param(
            ["1.0.0", "2.0.0"],
            "<2",
            ["python1.0"],
            {"1.0.0": pathlib.Path("/path/python1.0")},
            id="with-range-less-than-2",
        ),
    ],
)
def test_getPythonExecutables(
    monkeypatch: pytest.MonkeyPatch,
    availableVersions: typing.List[str],
    range_: typing.Optional[str],
    executables: typing.List[str],
    expectedExecutables: typing.Dict[str, str],
) -> None:
    repoData: typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]] = {"python": {}}

    for version in availableVersions:
        repoData["python"][version] = {"version": version}

    repo = typing.cast(
        rez.package_repository.PackageRepository,
        rez.package_repository.create_memory_package_repository(repoData),
    )

    with monkeypatch.context() as context:
        context.setitem(
            rez.package_repository.package_repository_manager.repositories,
            f"memory@{repo.location}",
            repo,
        )

        context.setattr(rez.config.config, "packages_path", [f"memory@{repo.location}"])

        with unittest.mock.patch(
            "rez.resolved_context.ResolvedContext.which"
        ) as mocked:
            mocked.side_effect = ["/path/" + exe if exe else "" for exe in executables]

            assert (
                rez_pip.rez.getPythonExecutables(range_, "python")
                == expectedExecutables
            )


def test_getPythonExecutables_isolation(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    """
    Test that getPythonExecutables is properly isolated. In other words, that
    system wide PATH does not affect ResolvedContext.which calls.

    If ResolvedContext.append_sys_path is not False, this test should fail.
    """
    packagePath = tmp_path / "package"
    fakePath = tmp_path / "fake"

    escapedPath = os.fspath(packagePath).replace("\\", "\\\\")
    # Create a fake python-1.0.0 package
    repoData: typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]] = {
        "python": {
            "1.0.0": {
                "version": "1.0.0",
                "commands": f"env.PATH.prepend('{escapedPath}')",
            },
        }
    }

    repo = typing.cast(
        rez.package_repository.PackageRepository,
        rez.package_repository.create_memory_package_repository(repoData),
    )

    with monkeypatch.context() as context:
        context.setitem(
            rez.package_repository.package_repository_manager.repositories,
            f"memory@{repo.location}",
            repo,
        )

        context.setattr(rez.config.config, "packages_path", [f"memory@{repo.location}"])

        ext = ".bat" if platform.system() == "Windows" else ""
        packagePath.mkdir()

        # Create the python executable in the python package.
        # Note that we use python1 here and we will use python1.0
        # in the fake python install so that the fake would have hiher priority
        # based on the algorithm in getPythonExecutables.
        with open(packagePath / f"python1{ext}", "w") as f:
            f.write("#FAKE EXECUTABLE")

        if platform.system() != "Windows":
            os.chmod(packagePath / f"python1{ext}", stat.S_IRUSR | stat.S_IXUSR)

        fakePath.mkdir()

        # Now create the fake executable.
        with open(fakePath / f"python1.0{ext}", "w") as f:
            f.write("#FAKE EXECUTABLE")

        if platform.system() != "Windows":
            os.chmod(fakePath / f"python1.0{ext}", stat.S_IRUSR | stat.S_IXUSR)

        def mockedFunc(self, *args, **kwargs):
            self.env.PATH.append(fakePath)

        # Mock append_system_paths to force our fake path into the
        # injected sys paths. If we don't do this, the fake path will not
        # be set in the ResolvedContext when append_system_paths is True.
        with unittest.mock.patch(
            "rez.rex.RexExecutor.append_system_paths",
            new=mockedFunc,
        ):
            assert rez_pip.rez.getPythonExecutables("1.0.0", "python") == {
                "1.0.0": packagePath / f"python1{ext}"
            }



import typing

import pytest
import packaging.version
import packaging.specifiers
import packaging.requirements
import rez.vendor.version.version
import rez.vendor.version.requirement

import rez_pip.utils


# @pytest.fixture
# def distPath():
#     cls.settings = {}
#     cls.dist_path = cls.data_path("pip", "installed_distributions")


def test_pythontDistributionNameToRez():
    """ """
    assert rez_pip.utils.pythontDistributionNameToRez("asd") == "asd"
    assert rez_pip.utils.pythontDistributionNameToRez("package-name") == "package_name"


@pytest.mark.parametrize(
    "version,expected",
    [
        ["1.0.0", "1.0.0"],
        ["0.9", "0.9"],
        ["1.0a1", "1.0.a1"],
        ["1.0.post1", "1.0.post1"],
        ["1.0.dev1", "1.0.dev1"],
        ["1.0+abc.7", "1.0-abc.7"],
        ["1!2.3.4", "2.3.4"],
    ],
)
def test_pythonDistributionVersionToRez(version: str, expected: str):
    assert rez_pip.utils.pythonDistributionVersionToRez(version) == expected


def test_test_pythonDistributionVersionToRez_raises():
    with pytest.raises(packaging.version.InvalidVersion):
        rez_pip.utils.pythonDistributionVersionToRez("2.0b1pl0")


@pytest.mark.parametrize(
    "pythonSpec,rezReq",
    [
        ["==1", "1+<1.1"],
        [">1", "1.1+"],
        ["<1", "<1"],
        [">=1", "1+"],
        ["<=1", "<1.1"],
        ["~=1.2", "1.2+<2"],
        ["~=1.2.3", "1.2.3+<1.3"],
        ["!=1", "<1|1.1+"],
        ["!=1.2", "<1.2|1.2.1+"],
        ["!=1.*", "<1|2+"],
        ["!=1.2.*", "<1.2|1.3+"],
        [">=1.2.a1", "1.2.a1+"],
        ["==1.*", "1"],
        [">=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4", "2.6+<3.0|3.3+<4"],
    ],
)
def test_pythonSpecifierToRezRequirement(pythonSpec: str, rezReq: str):
    pythonSpecObj = packaging.specifiers.SpecifierSet(pythonSpec)
    assert rez_pip.utils.pythonSpecifierToRezRequirement(
        pythonSpecObj
    ) == rez.vendor.version.version.VersionRange(rezReq)


def test_pythonSpecifierToRezRequirement_raises():
    with pytest.raises(ValueError):
        rez_pip.utils.pythonSpecifierToRezRequirement(
            packaging.specifiers.SpecifierSet("<2,>3")
        )


@pytest.mark.parametrize(
    "pythonReq,rezReq",
    [
        ["package>1", "package-1.1+"],
        ["package", "package"],
        ["package[extra]", "package"],
    ],
)
def test_packaging_req_to_rez_req(pythonReq: str, rezReq: str):
    assert rez_pip.utils.pythonReqToRezReq(
        packaging.requirements.Requirement(pythonReq)
    ) == rez.vendor.version.requirement.Requirement(rezReq)


# def test_is_pure_python_package(self):
#     """ """
#     dpath = rez.vendor.distlib.database.DistributionPath([self.dist_path])
#     dist = list(dpath.get_distributions())[0]

#     self.assertTrue(rez.utils.pip.is_pure_python_package(dist))


# def test_is_entry_points_scripts_package(self):
#     """ """
#     dpath = rez.vendor.distlib.database.DistributionPath([self.dist_path])
#     dist = list(dpath.get_distributions())[0]
#     self.assertFalse(rez.utils.pip.is_entry_points_scripts_package(dist))


# def test_convert_distlib_to_setuptools_wrong(self):
#     """ """
#     dpath = rez.vendor.distlib.database.DistributionPath([self.dist_path])
#     dist = list(dpath.get_distributions())[0]
#     dist.key = "random-unexisting-package"

#     self.assertEqual(rez.utils.pip.convert_distlib_to_setuptools(dist), None)


@pytest.mark.parametrize(
    "marker,expected",
    [
        ['implementation_name == "cpython"', ["python"]],
        ['implementation_version == "3.4.0"', ["python"]],
        ['platform_python_implementation == "Jython"', ["python"]],
        ['platform.python_implementation == "Jython"', ["python"]],
        ['python_implementation == "Jython"', ["python"]],
        ['sys_platform == "linux2"', ["platform"]],
        ['sys.platform == "linux2"', ["platform"]],
        ['os_name == "linux2"', ["platform"]],
        ['os.name == "linux2"', ["platform"]],
        ['platform_machine == "x86_64"', ["arch"]],
        ['platform.machine == "x86_64"', ["arch"]],
        ['platform_version == "#1 SMP Fri Apr 25 13:07:35 EDT 2014"', ["platform"]],
        ['platform.version == "#1 SMP Fri Apr 25 13:07:35 EDT 2014"', ["platform"]],
        ['platform_system == "Linux"', ["platform"]],
        ['platform_release == "5.2.8-arch1-1-ARCH"', ["platform"]],
        ['python_version == "3.7"', ["python"]],
        ['python_full_version == "3.7.4"', ["python"]],
    ],
)
def test_convertMarker(marker: str, expected: typing.List[str]):
    assert rez_pip.utils.convertMarker(marker) == expected


@pytest.mark.parametrize(
    "requirement,expected,conditional_extras",
    [
        ["packageA", [packaging.requirements.Requirement("packageA")], [None]],
        [
            "mypkg ; extra == 'dev'",
            [packaging.requirements.Requirement("mypkg")],
            [set(["dev"])],
        ],
        [
            'win-inet-pton ; (sys_platform == "win32" and python_version == "2.7") and extra == \'socks\'',
            [
                packaging.requirements.Requirement(
                    'win-inet-pton; (sys_platform == "win32" and python_version == "2.7")'
                )
            ],
            [set(["socks"])],
        ],
        # PySocks (!=1.5.7,<2.0,>=1.5.6) ; extra == 'socks'
        [
            "PySocks (!=1.5.7,<2.0,>=1.5.6) ; extra == 'socks'",
            [packaging.requirements.Requirement("PySocks!=1.5.7,<2.0,>=1.5.6")],
            [set(["socks"])],
        ],
        # certifi ; extra == 'secure'
        [
            "certifi ; extra == 'secure'",
            [packaging.requirements.Requirement("certifi")],
            [set(["secure"])],
        ],
        # coverage (>=4.4)
        [
            "coverage (>=4.4)",
            [packaging.requirements.Requirement("coverage (>=4.4)")],
            [None],
        ],
        # colorama ; sys_platform == "win32"
        [
            'colorama ; sys_platform == "win32"',
            [packaging.requirements.Requirement('colorama ; sys_platform == "win32"')],
            [None],
        ],
        # pathlib2-2.3.4: {u'environment': u'python_version<"3.5"', u'requires': [u'scandir']}, {u'requires': [u'six']}
        [
            {"environment": 'python_version<"3.5"', "requires": ["scandir"]},
            [packaging.requirements.Requirement('scandir; python_version < "3.5"')],
            [None],
        ],
        # bleach-3.1.0: {u'requires': [u'six (>=1.9.0)', u'webencodings']}
        [
            {"requires": ["six (>=1.9.0)", "webencodings"]},
            [
                packaging.requirements.Requirement("six (>=1.9.0)"),
                packaging.requirements.Requirement("webencodings"),
            ],
            [None, None],
        ],
        [
            {"requires": ["six (>=1.9.0)", 'webencodings; sys_platform == "win32"']},
            [
                packaging.requirements.Requirement("six (>=1.9.0)"),
                packaging.requirements.Requirement(
                    'webencodings; sys_platform == "win32"'
                ),
            ],
            [None, None],
        ],
        [
            {"requires": ["packageA"], "extra": "doc"},
            [packaging.requirements.Requirement("packageA")],
            [set(["doc"])],
        ],
        [
            "mypkg ; extra == 'dev' or extra == 'doc'",
            [packaging.requirements.Requirement("mypkg")],
            [set(["dev", "doc"])],
        ],
        [
            'mypkg ; extra == "dev" and sys_platform == "win32"',
            [packaging.requirements.Requirement('mypkg; sys_platform == "win32"')],
            [set(["dev"])],
        ],
        [
            'mypkg ; sys_platform == "win32" and extra == "test"',
            [packaging.requirements.Requirement('mypkg; sys_platform == "win32"')],
            [set(["test"])],
        ],
    ],
)
def test_normalizeRequirement(
    requirement: typing.Union[str, typing.Dict[str, typing.Any]],
    expected: typing.List[packaging.requirements.Requirement],
    conditional_extras: typing.List[typing.Optional[typing.Set[str]]],
):
    result = rez_pip.utils.normalizeRequirement(requirement)
    assert [str(req) for req in result] == [str(req) for req in expected]
    for index, req in enumerate(result):
        assert req.conditional_extras == conditional_extras[index]



import os
import sys
import typing
import hashlib
import pathlib

if sys.version_info[:2] < (3, 8):
    import mock
else:
    from unittest import mock

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import pytest
import aiohttp

import rez_pip.pip
import rez_pip.download


@pytest.fixture(scope="module", autouse=True)
def rezPipVersion():
    with mock.patch.object(importlib_metadata, "version", return_value="1.2.3.4.5"):
        yield


@pytest.mark.parametrize(
    "packages",
    [
        {"package-a": "package-a data"},
        {"package-a": "package-a data", "package-b": "package-b data"},
    ],
    ids=["single-package", "multiple-packages"],
)
def test_download(packages: typing.Dict[str, str], tmp_path: pathlib.Path):
    sideEffects = tuple()
    for content in packages.values():
        mockedContent = mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content.encode("utf-8"),
                None,
            ]
        ]

        sideEffects += (
            mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet = mock.AsyncMock()
    mockedGet.__aenter__.side_effect = sideEffects

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet

        wheels = rez_pip.download.downloadPackages(
            [
                rez_pip.pip.PackageInfo(
                    metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                    download_info=rez_pip.pip.DownloadInfo(
                        url=f"https://example.com/{package}.whl",
                        archive_info=rez_pip.pip.ArchiveInfo("hash", {}),
                    ),
                    is_direct=True,
                    requested=True,
                )
                for package in packages
            ],
            os.fspath(tmp_path),
        )

    assert sorted(wheels) == sorted(
        [os.fspath(tmp_path / f"{package}.whl") for package in packages]
    )

    for wheel in wheels:
        with open(wheel, "r") as fd:
            content = fd.read()
        assert packages[os.path.basename(wheel).split(".")[0]] == content

    assert mocked.call_args_list == [
        mock.call(
            f"https://example.com/{package}.whl",
            headers={
                "Content-Type": "application/octet-stream",
                "User-Agent": "rez-pip/1.2.3.4.5",
            },
        )
        for package in packages
    ]


def test_download_multiple_packages_with_failure(tmp_path: pathlib.Path):
    mockedContent = mock.MagicMock()
    mockedContent.return_value.__aiter__.return_value = [
        [
            b"package-a data",
            None,
        ]
    ]

    mockedGet = mock.AsyncMock()
    mockedGet.__aenter__.side_effect = (
        mock.Mock(
            headers={"content-length": 100},
            status=200,
            content=mock.Mock(iter_chunks=mockedContent),
        ),
        mock.Mock(
            headers={"content-length": 100},
            status=400,
            reason="Expected to fail",
            request_info={"key": "here"},
            content=mock.Mock(iter_chunks=mockedContent),
        ),
    )

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet
        with pytest.raises(RuntimeError):
            rez_pip.download.downloadPackages(
                [
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(
                            name="package-a", version="1.0.0"
                        ),
                        download_info=rez_pip.pip.DownloadInfo(
                            url="https://example.com/package-a",
                            archive_info=rez_pip.pip.ArchiveInfo("hash-a", {}),
                        ),
                        is_direct=True,
                        requested=True,
                    ),
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(
                            name="package-b", version="1.0.0"
                        ),
                        download_info=rez_pip.pip.DownloadInfo(
                            url="https://example.com/package-b",
                            archive_info=rez_pip.pip.ArchiveInfo("hash-b", {}),
                        ),
                        is_direct=True,
                        requested=True,
                    ),
                ],
                os.fspath(tmp_path),
            )

        # Check that package-a was downloaded even if even if package-b failed.
        with open(tmp_path / "package-a", "r") as fd:
            assert fd.read() == "package-a data"

        assert mocked.call_args_list == [
            mock.call(
                "https://example.com/package-a",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            mock.call(
                "https://example.com/package-b",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]


def test_download_reuse_if_same_hash(tmp_path: pathlib.Path):
    """Test that wheels are re-used if the sha256 matches"""
    sideEffects = tuple()
    packages = []

    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        packages.append(
            rez_pip.pip.PackageInfo(
                metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                download_info=rez_pip.pip.DownloadInfo(
                    url=f"https://example.com/{package}.whl",
                    archive_info=rez_pip.pip.ArchiveInfo(
                        "hash-a", {"sha256": hash.hexdigest()}
                    ),
                ),
                is_direct=True,
                requested=True,
            )
        )

        mockedContent = mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet1 = mock.AsyncMock()
    mockedGet1.__aenter__.side_effect = sideEffects

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet1

        rez_pip.download.downloadPackages(packages, str(tmp_path))

        assert mocked.call_args_list == [
            mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    packages = []
    # package-b will be re-used
    for package in ["package-c", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        packages.append(
            rez_pip.pip.PackageInfo(
                metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                download_info=rez_pip.pip.DownloadInfo(
                    url=f"https://example.com/{package}.whl",
                    archive_info=rez_pip.pip.ArchiveInfo(
                        "hash-a", {"sha256": hash.hexdigest()}
                    ),
                ),
                is_direct=True,
                requested=True,
            )
        )

        mockedContent = mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet2 = mock.AsyncMock()
    mockedGet2.__aenter__.side_effect = sideEffects

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet2

        wheels = rez_pip.download.downloadPackages(packages, str(tmp_path))

        assert mocked.call_args_list == [
            mock.call(
                "https://example.com/package-c.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    assert sorted(wheels) == [
        os.fspath(tmp_path / f"{package}.whl") for package in ["package-b", "package-c"]
    ]


def test_download_redownload_if_hash_changes(tmp_path: pathlib.Path):
    """Test that wheels are re-used if the sha256 matches"""
    sideEffects = tuple()
    packages = []

    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        packages.append(
            rez_pip.pip.PackageInfo(
                metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                download_info=rez_pip.pip.DownloadInfo(
                    url=f"https://example.com/{package}.whl",
                    archive_info=rez_pip.pip.ArchiveInfo(
                        "hash-a", {"sha256": hash.hexdigest()}
                    ),
                ),
                is_direct=True,
                requested=True,
            )
        )

        mockedContent = mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet1 = mock.AsyncMock()
    mockedGet1.__aenter__.side_effect = sideEffects

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet1

        rez_pip.download.downloadPackages(packages, str(tmp_path))

        assert mocked.call_args_list == [
            mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    packages = []
    # package-b will be re-used
    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        packages.append(
            rez_pip.pip.PackageInfo(
                metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                download_info=rez_pip.pip.DownloadInfo(
                    url=f"https://example.com/{package}.whl",
                    archive_info=rez_pip.pip.ArchiveInfo(
                        #
                        # Bad sha256. This will trigger a new download
                        #
                        "hash-a",
                        {"sha256": "asd"},
                    ),
                ),
                is_direct=True,
                requested=True,
            )
        )

        mockedContent = mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet2 = mock.AsyncMock()
    mockedGet2.__aenter__.side_effect = sideEffects

    with mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet2

        wheels = rez_pip.download.downloadPackages(packages, str(tmp_path))

        assert mocked.call_args_list == [
            mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    assert sorted(wheels) == [
        os.fspath(tmp_path / f"{package}.whl") for package in ["package-a", "package-b"]
    ]



import sys
import typing
import logging
import dataclasses

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import rez.system
import packaging.version
import packaging.specifiers
import packaging.requirements
import rez.vendor.version.version
import rez.vendor.version.requirement

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class RequirementsDict:
    requires: typing.List[str]
    variant_requires: typing.List[str]
    metadata: typing.Dict[str, typing.Any]


def pythontDistributionNameToRez(name: str) -> str:
    """Convert a distribution name to a rez compatible name.

    The rez package name can't be simply set to the dist name, because some
    pip packages have hyphen in the name. In rez this is not a valid package
    name (it would be interpreted as the start of the version).
    Example: my-pkg-1.2 is 'my', version 'pkg-1.2'.

    :param name: Distribution name to convert.
    :returns: Rez-compatible package name.
    """
    return name.replace("-", "_")


def pythonDistributionVersionToRez(version: str) -> str:
    """Convert a distribution version to a rez compatible version.

    The python version schema specification isn't 100% compatible with rez.

    1: version epochs (they make no sense to rez, so they'd just get stripped
       of the leading N!;
    2: python versions are case insensitive, so they should probably be
       lowercased when converted to a rez version.
    3: local versions are also not compatible with rez

    The canonical public version identifiers MUST comply with the following scheme:
    [N!]N(.N)*[{a|b|rc}N][.postN][.devN]

    Epoch segment: N! - skip
    Release segment: N(.N)* 0 as is
    Pre-release segment: {a|b|c|rc|alpha|beta|pre|preview}N - always lowercase
    Post-release segment: .{post|rev|r}N - always lowercase
    Development release segment: .devN - always lowercase

    Local version identifiers MUST comply with the following scheme:
    <public version identifier>[+<local version label>] - use - instead of +

    :param version: The distribution version to be converted.
    :returns: Rez-compatible equivalent version string.
    :raises packaging.version.InvalidVersion: when a PEP440 incompatible version is detected.

    .. _Core utilities for Python packages:
        https://packaging.pypa.io/en/latest/version/
    """
    pkgVersion = packaging.version.parse(version)

    # the components of the release segment excluding epoch or any
    # prerelease/development/postrelease suffixes
    rezVersion = ".".join(str(i) for i in pkgVersion.release)

    if pkgVersion.is_prerelease and pkgVersion.pre:
        # Additional check is necessary because dev releases are also considered
        # prereleases pair of the prerelease phase (the string "a", "b", or "rc")
        # and the prerelease number the following conversions take place:
        # * a --> a
        # * alpha --> a
        # * b --> b
        # * beta --> b
        # * c --> c
        # * rc --> rc
        # * pre --> rc
        # * preview --> rc
        #
        phase, number = pkgVersion.pre
        rezVersion += "." + phase + str(number)

    if pkgVersion.is_postrelease:
        # this attribute will be the postrelease number (an integer)
        # the following conversions take place:
        # * post --> post
        # * rev --> post
        # * r --> post
        #
        rezVersion += ".post" + str(pkgVersion.post)

    if pkgVersion.is_devrelease:
        # this attribute will be the development release number (an integer)
        rezVersion += ".dev" + str(pkgVersion.dev)

    if pkgVersion.local:
        # representation of the local version portion is any
        # the following conversions take place:
        # 1.0[+ubuntu-1] --> 1.0[-ubuntu.1]
        rezVersion += "-" + pkgVersion.local

    return rezVersion


def pythonSpecifierToRezRequirement(
    specifier: packaging.specifiers.SpecifierSet,
) -> rez.vendor.version.version.VersionRange:
    """Convert PEP440 version specifier to rez equivalent.

    See https://www.python.org/dev/peps/pep-0440/#version-specifiers

    Note that version numbers in the specifier are converted to rez equivalents
    at the same time. Thus a specifier like '<1.ALPHA2' becomes '<1.a2'.

    Note that the conversion is not necessarily exact - there are cases in
    PEP440 that have no equivalent in rez versioning. Most of these are
    specifiers that involve pre/post releases, which don't exist in rez (or
    rather, they do exist in the sense that '1.0.post1' is a valid rez version
    number, but it has no special meaning).

    Note also that the specifier is being converted into rez format, but in a
    way that still expresses how _pip_ interprets the specifier. For example,
    '==1' is a valid version range in rez, but '==1' has a different meaning to
    pip than it does to rez ('1.0' matches '==1' in pip, but not in rez). This
    is why '==1' is converted to '1+<1.1' in rez, rather than '==1'.

    Example conversions:

        |   PEP440    |     rez     |
        |-------------|-------------|
        | ==1         | 1+<1.1      |
        | ==1.*       | 1           |
        | >1          | 1.1+        |
        | <1          | <1          |
        | >=1         | 1+          |
        | <=1         | <1.1        |
        | ~=1.2       | 1.2+<2      |
        | ~=1.2.3     | 1.2.3+<1.3  |
        | !=1         | <1|1.1+     |
        | !=1.2       | <1.2|1.2.1+ |
        | !=1.*       | <1|2+       |
        | !=1.2.*     | <1.2|1.3+   |

    :param specifier: Pip specifier.

    `VersionRange`: Equivalent rez version range.
    """

    def is_release(rezVer: str) -> bool:
        parts = rezVer.split(".")
        try:
            _ = int(parts[-1])  # noqa
            return True
        except:
            return False

    # 1 --> 2; 1.2 --> 1.3; 1.a2 -> 1.0
    def next_ver(rezVer: str) -> str:
        parts = rezVer.split(".")
        if is_release(rezVer):
            parts = parts[:-1] + [str(int(parts[-1]) + 1)]
        else:
            parts = parts[:-1] + ["0"]
        return ".".join(parts)

    # 1 --> 1.1; 1.2 --> 1.2.1; 1.a2 --> 1.0
    def adjacent_ver(rezVer: str) -> str:
        if is_release(rezVer):
            return rezVer + ".1"
        else:
            parts = rezVer.split(".")
            parts = parts[:-1] + ["0"]
            return ".".join(parts)

    def convert_spec(spec: packaging.specifiers.Specifier) -> str:
        def parsed_rez_ver() -> str:
            v = spec.version.replace(".*", "")
            return pythonDistributionVersionToRez(v)

        def fmt(txt: str) -> str:
            v = parsed_rez_ver()
            vnext = next_ver(v)
            vadj = adjacent_ver(v)
            return txt.format(V=v, VNEXT=vnext, VADJ=vadj)

        # ==1.* --> 1
        if spec.operator == "==" and spec.version.endswith(".*"):
            return fmt("{V}")

        # ==1 --> 1+<1.1
        if spec.operator == "==":
            return fmt("{V}+<{VADJ}")

        # >=1 --> 1+
        if spec.operator == ">=":
            return fmt("{V}+")

        # >1 --> 1.1+
        if spec.operator == ">":
            return fmt("{VADJ}+")

        # <= 1 --> <1.1
        if spec.operator == "<=":
            return fmt("<{VADJ}")

        # <1 --> <1
        if spec.operator == "<":
            return fmt("<{V}")

        # ~=1.2 --> 1.2+<2; ~=1.2.3 --> 1.2.3+<1.3
        if spec.operator == "~=":
            v = rez.vendor.version.version.Version(parsed_rez_ver())
            v = v.trim(len(v) - 1)
            v_next = next_ver(str(v))
            return fmt("{V}+<" + v_next)

        # !=1.* --> <1|2+; !=1.2.* --> <1.2|1.3+
        if spec.operator == "!=" and spec.version.endswith(".*"):
            return fmt("<{V}|{VNEXT}+")

        # !=1 --> <1|1.1+; !=1.2 --> <1.2|1.2.1+
        if spec.operator == "!=":
            return fmt("<{V}|{VADJ}+")

        raise ValueError(
            f"Don't know how to convert PEP440 specifier {specifier!r} "
            "into rez equivalent"
        )

    # convert each spec into rez equivalent
    ranges = list(map(convert_spec, specifier))

    # AND together ranges
    total_range = rez.vendor.version.version.VersionRange(ranges[0])

    for range_ in ranges[1:]:
        range_ = rez.vendor.version.version.VersionRange(range_)
        total_range = total_range.intersection(range_)

        if total_range is None:
            raise ValueError(
                f"PEP440 specifier {specifier} converts to a non-intersecting rez "
                "version range"
            )

    return total_range


def pythonReqToRezReq(
    pythonReq: packaging.requirements.Requirement,
) -> rez.vendor.version.requirement.Requirement:
    """Convert packaging requirement object to equivalent rez requirement.

    Note that environment markers are ignored.

    :param pythonReq: Python requirement.
    :returns: Equivalent rez requirement object.
    """
    if pythonReq.extras:
        _LOG.warning(
            f"Ignoring extras requested on {pythonReq!r} - this is not yet supported"
        )

    req = pythontDistributionNameToRez(pythonReq.name)

    if pythonReq.specifier:
        range_ = pythonSpecifierToRezRequirement(pythonReq.specifier)
        req += "-" + str(range_)

    return rez.vendor.version.requirement.Requirement(req)


class CustomPyPackagingRequirement(packaging.requirements.Requirement):
    conditional_extras: typing.Optional[typing.Set[str]]


def normalizeRequirement(
    requirement: typing.Union[str, typing.Dict[typing.Any, typing.Any]]
) -> typing.List[CustomPyPackagingRequirement]:
    """Normalize a package requirement.

    Requirements from distlib packages can be a mix of string- or dict- based
    formats, as shown here:
    * https://www.python.org/dev/peps/pep-0508/#environment-markers
    * https://legacy.python.org/dev/peps/pep-0426/#environment-markers

    There's another confusing case that this code deals with. Consider these two requirements:

        # means: reportlab is a requirement of this package when the 'pdf' extra is requested
        Requires-Dist: reportlab; extra == 'pdf'
        means: this package requires libexample, with its 'test' extras
        Requires-Dist: libexample[test]

    See https://packaging.python.org/specifications/core-metadata/#provides-extra-multiple-use

    The packaging lib doesn't do a good job of expressing this - the first form
    of extras use just gets embedded in the environment marker. This function
    parses the extra from the marker, and stores it onto the resulting
    `packaging.requirements.Requirement` object in a 'conditional_extras' attribute. It also
    removes the extra from the marker (otherwise the marker cannot evaluate).
    Even though you can specify `environment` in `packaging.markers.Marker.evaluate`,
    you can only supply a single 'extra' key in the env, so this can't be used
    to correctly evaluate if multiple extras were requested.

    :param requirement: Requirement, for eg from `importlib.metadata.Distribution.requires`.
    :returns: Normalized requirements. Note that a list is returned, because the PEP426 format can define
        multiple requirements.
    """

    def reconstruct(
        req: CustomPyPackagingRequirement,
        marker_str: typing.Optional[str] = None,
        conditional_extras: typing.Union[typing.Set[str], None] = None,
    ) -> CustomPyPackagingRequirement:
        new_req_str = req.name

        if req.specifier:
            new_req_str += " (%s)" % str(req.specifier)

        if marker_str is None and req.marker:
            marker_str = str(req.marker)

        if marker_str:
            new_req_str += " ; " + marker_str

        new_req = CustomPyPackagingRequirement(new_req_str)
        new_req.conditional_extras = conditional_extras
        return new_req

    # PEP426 dict syntax
    # So only metadata that are of version 2.0 will be in dict. The other versions
    # (1.0, 1.1, 1.2, 2.1) will be strings.
    if isinstance(requirement, dict):
        result: typing.List[CustomPyPackagingRequirement] = []
        requires = requirement["requires"]
        extra = requirement.get("extra")
        marker_str = requirement.get("environment")

        # conditional extra, equivalent to: 'foo ; extra = "doc"'
        if extra:
            conditional_extras1 = set([extra])
        else:
            conditional_extras1 = None

        for req_str in requires:
            req = CustomPyPackagingRequirement(req_str)
            new_req = reconstruct(req, marker_str, conditional_extras1)
            result.append(new_req)

        return result

    # string-based syntax
    req = CustomPyPackagingRequirement(requirement)

    # detect case: "mypkg ; extra == 'dev'"
    # note: packaging lib already delimits with whitespace
    marker_str = str(req.marker)
    marker_parts = marker_str.split()

    # already in PEP508, packaging lib- friendly format
    if "extra" not in marker_parts:
        req.conditional_extras = None
        return [req]

    # Parse conditional extras out of marker
    conditional_extras: typing.Set[str] = set()
    marker_str = marker_str.replace(" and ", " \nand ")
    marker_str = marker_str.replace(" or ", " \nor ")
    lines = marker_str.split("\n")
    lines = [x.strip() for x in lines]
    new_marker_lines: typing.List[str] = []

    for line in lines:
        if "extra" in line.split():
            extra = line.split()[-1]
            extra = extra.replace('"', "")
            extra = extra.replace("'", "")
            conditional_extras.add(extra)
        else:
            new_marker_lines.append(line)

    # reconstruct requirement in new syntax
    if new_marker_lines:
        new_marker_parts = " ".join(new_marker_lines).split()
        if new_marker_parts[0] in ("and", "or"):
            new_marker_parts = new_marker_parts[1:]
        new_marker_str = " ".join(new_marker_parts)
    else:
        new_marker_str = ""

    new_req = reconstruct(req, new_marker_str, conditional_extras)
    return [new_req]


def convertMarker(marker: str) -> typing.List[str]:
    """Get the system requirements that an environment marker introduces.

    Consider:
        'foo (>1.2) ; python_version == "3" and platform_machine == "x86_64"'

    This example would cause a requirement on python, platform, and arch
    (platform as a consequence of requirement on arch).

    See:
    * vendor/packaging/markers.py:line=76
    * https://www.python.org/dev/peps/pep-0508/#id23

    :param marker: Environment marker string, eg 'python_version == "3"'.
    :returns: System requirements (unversioned).
    """
    _py = "python"
    _plat = "platform"
    _arch = "arch"

    sys_requires_lookup = {
        # TODO There is no way to associate a python version with its implementation
        # currently (ie CPython etc). When we have "package features", we may be
        # able to manage this; ignore for now
        "implementation_name": [_py],  # PEP-0508
        "implementation_version": [_py],  # PEP-0508
        "platform_python_implementation": [_py],  # PEP-0508
        "platform.python_implementation": [_py],  # PEP-0345
        "python_implementation": [
            _py
        ],  # setuptools legacy. Same as platform_python_implementation
        "sys.platform": [_plat],  # PEP-0345
        "sys_platform": [_plat],  # PEP-0508
        # note that this maps to python's os.name, which does not mean distro
        # (as 'os' does in rez). See https://docs.python.org/2/library/os.html#os.name
        "os.name": [_plat],  # PEP-0345
        "os_name": [_plat],  # PEP-0508
        "platform.machine": [_arch],  # PEP-0345
        "platform_machine": [_arch],  # PEP-0508
        # TODO hmm, we never variant on plat version, let's leave this for now...
        "platform.version": [_plat],  # PEP-0345
        "platform_version": [_plat],  # PEP-0508
        # somewhat ambiguous cases
        "platform_system": [_plat],  # PEP-0508
        "platform_release": [_plat],  # PEP-0508
        "python_version": [_py],  # PEP-0508
        "python_full_version": [_py],  # PEP-0508
    }

    sys_requires: typing.Set[str] = set()

    # note: packaging lib already delimits with whitespace
    marker_parts = marker.split()

    for varname, sys_reqs in sys_requires_lookup.items():
        if varname in marker_parts:
            sys_requires.update(sys_reqs)

    return list(sys_requires)


def getRezRequirements(
    installedDist: importlib_metadata.Distribution,
    pythonVersion: rez.vendor.version.version.Version,
    isPure: bool,
    nameCasings: typing.Optional[typing.List[str]] = None,
) -> RequirementsDict:
    """Get requirements of the given dist, in rez-compatible format.

    Example result:
        {
            "requires": ["foo-1.2+<2"],
            "variant_requires": ["future", "python-2.7"],
            "metadata": {
                # metadata pertinent to rez
                ...
            }
        }

    Each requirement has had its package name converted to the rez equivalent.

    The 'variant_requires' key contains requirements specific to the current
    variant.

    TODO: Currently there is no way to reflect extras that may have been chosen

    for this pip package. We need to wait for rez "package features" before this
    will be possible. You probably shouldn't use extras presently.

    :param installedDist: Distribution to convert.
    :param pythonVersion: Python version used to perform the installation.
    :param nameCasings: A list of pip package names in their correct
        casings (eg, 'Foo' rather than 'foo'). Any requirement whose name
        case-insensitive-matches a name in this list, is set to that name.
        This is needed because pip package names are case insensitive, but
        rez is case-sensitive. So a package may list a requirement for package
        'foo', when in fact the package that pip has downloaded is called 'Foo'.
        Be sure to provide names in PIP format, not REZ format (the pip package
        'foo-bah' will be converted to 'foo_bah' in rez).
    :returns: See example above.
    """
    _system = rez.system.System()
    result_requires: typing.List[str] = []
    result_variant_requires: typing.List[str] = []

    # create cased names lookup
    name_mapping = dict((x.lower(), x) for x in (nameCasings or []))

    # requirements such as platform, arch, os, and python
    sys_requires: typing.Set[str] = set()

    # entry_points scripts are platform and arch specific executables generated by
    # python build frontends during install
    has_entry_points_scripts = bool(installedDist.entry_points)

    # assume package is platform- and arch- specific if it isn't pure python
    if not isPure or has_entry_points_scripts:
        sys_requires.update(["platform", "arch"])

    # evaluate wrt python version, which may not be the current interpreter version

    # JC: TODO: Use rez to resolve the ful npython version based on the provided version?
    marker_env = {
        "python_full_version": str(pythonVersion),
        "python_version": str(pythonVersion.trim(2)),
        "implementation_version": str(pythonVersion),
    }

    # Note: This is supposed to give a requirements list that has already been
    # filtered down based on the extras requested at install time, and on any
    # environment markers present. However, this is not working in distlib. The
    # package gets assigned a LegacyMetadata metadata object (only if a package metadata
    # version is not equal to 2.0) and in that code path, this filtering
    # doesn't happen.
    #
    # See: vendor/distlib/metadata.py#line-892
    #
    requires = installedDist.requires

    # filter requirements
    for req_ in requires or []:
        reqs = normalizeRequirement(req_)

        for req in reqs:
            # skip if env marker is present and doesn't evaluate
            if req.marker and not req.marker.evaluate(environment=marker_env):
                continue

            # skip if req is conditional on extras that weren't requested
            if req.conditional_extras and not (
                set(installedDist.metadata["Provides-Extra"] or [])
                & req.conditional_extras
            ):
                continue

            if req.conditional_extras:
                _LOG.warning(
                    f"Skipping requirement {req!r} - conditional requirements are "
                    "not yet supported"
                )
                continue

            # Inspect marker(s) to see if this requirement should be varianted.
            # Markers may also cause other system requirements to be added to
            # the variant.
            #
            to_variant = False

            if req.marker:
                marker_reqs = convertMarker(str(req.marker))

                if marker_reqs:
                    sys_requires.update(marker_reqs)
                    to_variant = True

            # remap the requirement name
            remapped = name_mapping.get(req.name.lower())
            if remapped:
                req.name = remapped

            # convert the requirement to rez equivalent
            rez_req = str(pythonReqToRezReq(req))

            if to_variant:
                result_variant_requires.append(rez_req)
            else:
                result_requires.append(rez_req)

    # prefix variant with system requirements
    sys_variant_requires: typing.List[str] = []

    if "platform" in sys_requires:
        sys_variant_requires.append(f"platform-{_system.platform}")

    # TODO: Support wheel tags. This is important to avoid having to rez-pip
    # for when isnt' not necessary.
    # https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/
    # So basically support manylinux and friends.
    if "arch" in sys_requires:
        sys_variant_requires.append(f"arch-{_system.arch}")

    if "os" in sys_requires:
        sys_variant_requires.append(f"os-{_system.os}")

    # JC: TODO: This is a modified version based on rez. It needs to be verified.
    if not isPure:
        # Add python variant requirement. Note that this is always MAJOR.MINOR,
        # because to do otherwise would mean analysing any present env markers.
        # This could become quite complicated, and could also result in strange
        # python version ranges in the variants.
        #
        sys_variant_requires.append("python-%s" % str(pythonVersion.trim(2)))
    else:
        requiresPython = installedDist.metadata["Requires-Python"]
        if requiresPython:
            result_requires.append(
                f"python-{pythonSpecifierToRezRequirement(packaging.specifiers.SpecifierSet(requiresPython))}"
            )
        else:
            # Some packages still don't provide Requires-Python. So make sure we
            # add python.
            # TODO: This is not not necessary for packages that don't have any python code...
            #       Thinking about cmake, ninja, etc. Though these are usually only requyired
            #       for building packages.
            result_requires.append("python")

    variant_requires = sys_variant_requires + result_variant_requires

    translation = {
        "python": requires,
        "rez": {"requires": result_requires, "variant": variant_requires},
    }
    _LOG.debug(f"{installedDist.name} requirements translation: {translation}")

    return RequirementsDict(
        requires=result_requires,
        variant_requires=variant_requires,
        metadata={"is_pure_python": isPure},  # TODO: This is probably useless.
    )



import os
import re
import platform
import subprocess

import pytest
import rez.packages
import rez.resolved_context
import rez.vendor.version.version


@pytest.mark.skipif(
    platform.system() == "Darwin" and not os.environ.get("CI"),
    reason="Skipping when running locally on macOS",
)
@pytest.mark.integration
def test_python_packages(pythonRezPackage: str, rezRepo: str):
    """Test that the python rez packages created by the pythonRezPackage fixture are functional"""

    package = rez.packages.get_package("python", pythonRezPackage, paths=[rezRepo])
    assert package

    ctx = rez.resolved_context.ResolvedContext(
        [package.qualified_name], package_paths=[rezRepo]
    )

    executable = f"python{str(package.version).split('.')[0]}"
    if platform.system() == "Windows":
        executable = "python.exe"

    code, stdout, _ = ctx.execute_shell(
        command=[executable, "--version"],
        block=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    assert code == 0
    assert re.findall(r"\d\.\d+\.\d+", stdout.decode("utf-8"), flags=re.MULTILINE)[
        0
    ] == str(package.version)

    code, stdout, _ = ctx.execute_shell(
        command=[executable, "-c", "import sys; print(sys.executable)"],
        block=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    expectedPath = os.path.join(
        rezRepo,
        "python",
        str(package.version),
        "python",
        "tools" if platform.system() == "Windows" else "bin",
        executable,
    )

    assert code == 0
    assert stdout.decode("utf-8").strip().lower() == expectedPath.lower()



import os
import sys
import json
import shutil
import typing
import logging
import argparse
import textwrap
import pathlib
import tempfile
import subprocess

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import rich
import rich.text
import rich.panel
import rich.markup
import rich.logging
import rez.vendor.version.version

import rez_pip.pip
import rez_pip.rez
import rez_pip.data
import rez_pip.install
import rez_pip.download
import rez_pip.exceptions

_LOG = logging.getLogger("rez_pip.cli")

__all__ = ["run"]


def __dir__() -> typing.List[str]:
    return __all__


def _parseArgs(
    args: typing.List[str],
) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
    parser = argparse.ArgumentParser(
        description="Ingest and convert python packages to rez packages.",
        prog=__package__.replace("_", "-"),
        add_help=False,
    )
    parser.add_argument("packages", nargs="*", help="Packages to install.")

    generalGroup = parser.add_argument_group(title="general options")
    generalGroup.add_argument(
        "-r",
        "--requirement",
        action="append",
        metavar="<file>",
        help="Install from the given requirements file. This option can be used multiple times.",
    )
    generalGroup.add_argument(
        "-c",
        "--constraint",
        action="append",
        metavar="<file>",
        help="Constrain versions using the given constraints file. This option can be used multiple times.",
    )
    generalGroup.add_argument(
        "-p",
        "--prefix",
        metavar="<path>",
        help="Custom repository path (can be any directory, even non rez repository path) (default: configured local_packages_path)",
    )
    generalGroup.add_argument(
        "--release",
        action="store_true",
        help="Release the converted packages (Default: configured release_packages_path)",
    )

    generalGroup.add_argument(
        "--python-version",
        # 3.7+ because that's what pip supports
        default="3.7+",
        metavar="<version>",
        help="Range of python versions. It can also be a single version, any valid rez version specifier or range or 'latest' (default: 3.7+)",
    )
    generalGroup.add_argument(
        "--pip",
        default=rez_pip.pip.getBundledPip(),
        metavar="<path>",
        help="Standalone pip (https://pip.pypa.io/en/stable/installation/#standalone-zip-application) (default: bundled).",
    )

    # Manually define just to keep the style consistent (capital letters, dot, etc.)
    generalGroup.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )

    debugGroup = parser.add_argument_group(title="debug options")
    debugGroup.add_argument(
        "-l",
        "--log-level",
        default="info",
        choices=["info", "debug", "warning", "error"],
        help="Logging level.",
    )

    debugGroup.add_argument(
        "--keep-tmp-dirs",
        action="store_true",
        help="Keep some temporary directories at the end of the process for further inspection.",
    )

    debugGroup.add_argument(
        "--debug-info",
        action="store_true",
        help="Print debug information that you can use when reporting an issue on GitHub.",
    )

    parser.usage = f"""

  %(prog)s [options] <package(s)>
  %(prog)s <package(s)> [-- [pip options]]
"""
    knownArgs = []
    pipArgs = []
    if "--" in args:
        # anything after -- will be passed as is to pip.
        splitIndex = args.index("--")
        knownArgs = args[:splitIndex]
        pipArgs = args[splitIndex + 1 :]
    else:
        knownArgs = args

    ownArgs = parser.parse_args(knownArgs)

    return ownArgs, pipArgs


def _validateArgs(args: argparse.Namespace) -> None:
    if not args.pip.endswith(".pyz"):
        raise rez_pip.exceptions.RezPipError(
            f"[bold red]{args.pip!r} does not look like a valid zipapp. A zipapp should end with '.pyz'.[/]\n\n"
            "  [grey74]Standalone pip documentation[/]: https://pip.pypa.io/en/stable/installation/#standalone-zip-application\n"
            "  [grey74]zipapp documentation[/]: https://docs.python.org/3/library/zipapp.html"
        )

    if not os.path.exists(args.pip):
        raise rez_pip.exceptions.RezPipError(rf"zipapp at {args.pip!r} does not exist")

    if not args.packages and not args.requirement:
        raise rez_pip.exceptions.RezPipError(
            "no packages were passed and --requirements was not used. At least one of them must be passed."
        )


def _run(args: argparse.Namespace, pipArgs: typing.List[str], pipWorkArea: str) -> None:
    pythonVersions = rez_pip.rez.getPythonExecutables(
        args.python_version, packageFamily="python"
    )

    for pythonVersion, pythonExecutable in pythonVersions.items():
        _LOG.info(
            f"[bold underline]Installing requested packages for Python {pythonVersion}"
        )

        wheelsDir = os.path.join(pipWorkArea, "wheels")
        os.makedirs(wheelsDir, exist_ok=True)

        # Suffix with the python version because we loop over multiple versions,
        # and package versions, content, etc can differ for each Python version.
        installedWheelsDir = os.path.join(pipWorkArea, "installed", pythonVersion)
        os.makedirs(installedWheelsDir, exist_ok=True)

        with rich.get_console().status(
            f"[bold]Resolving dependencies for {rich.markup.escape(', '.join(args.packages))} (python-{pythonVersion})"
        ):
            packages = rez_pip.pip.getPackages(
                args.packages,
                args.pip,
                pythonVersion,
                os.fspath(pythonExecutable),
                args.requirement or [],
                args.constraint or [],
                pipArgs,
            )

        _LOG.info(f"Resolved {len(packages)} dependencies for python {pythonVersion}")

        # TODO: Should we postpone downloading to the last minute if we can?
        _LOG.info("[bold]Downloading...")
        wheels = rez_pip.download.downloadPackages(packages, wheelsDir)
        _LOG.info(f"[bold]Downloaded {len(wheels)} wheels")

        dists: typing.Dict[importlib_metadata.Distribution, bool] = {}

        with rich.get_console().status(
            f"[bold]Installing wheels into {installedWheelsDir!r}"
        ):
            for package, wheel in zip(packages, wheels):
                _LOG.info(f"[bold]Installing {package.name}-{package.version} wheel")
                dist, isPure = rez_pip.install.installWheel(
                    package, pathlib.Path(wheel), installedWheelsDir
                )

                dists[dist] = isPure

        distNames = [dist.name for dist in dists.keys()]

        with rich.get_console().status("[bold]Creating rez packages..."):
            for dist, package in zip(dists, packages):
                isPure = dists[dist]
                rez_pip.rez.createPackage(
                    dist,
                    isPure,
                    rez.vendor.version.version.Version(pythonVersion),
                    distNames,
                    installedWheelsDir,
                    wheelURL=package.download_info.url,
                    prefix=args.prefix,
                    release=args.release,
                )


def _debug(
    args: argparse.Namespace, console: rich.console.Console = rich.get_console()
) -> None:
    """Print debug information"""
    prefix = "  "
    console.print(
        f"[bold]rez-pip version[/]: {importlib_metadata.version(__package__)}"
    )

    console.print(f"[bold]rez version[/]: {importlib_metadata.version('rez')}")

    console.print(f"[bold]python version[/]: {sys.version}", highlight=False)
    console.print(f"[bold]python executable[/]: {sys.executable}", highlight=False)

    pip = args.pip or rez_pip.pip.getBundledPip()

    console.print(f"[bold]pip[/]: {pip}", highlight=False)

    completedProcess = subprocess.run(
        [sys.executable, pip, "--version"],
        stdout=subprocess.PIPE,
        text=True,
    )

    console.print(
        f"[bold]pip version[/]: {completedProcess.stdout.strip()}", highlight=False
    )

    completedProcess = subprocess.run(
        [sys.executable, pip, "config", "debug"],
        stdout=subprocess.PIPE,
        text=True,
    )

    console.print("[bold]rez-pip provided arguments[/]:")
    print(textwrap.indent(json.dumps(vars(args), indent=4), prefix))

    console.print(f"[bold]pip config debug[/]:", highlight=False)
    print(textwrap.indent(completedProcess.stdout.strip(), "  "))

    completedProcess = subprocess.run(
        [sys.executable, pip, "config", "list"],
        stdout=subprocess.PIPE,
        text=True,
    )

    console.print(f"[bold]pip config list[/]:", highlight=False)
    print(
        textwrap.indent(completedProcess.stdout.strip(), prefix)
        or f"{prefix}Returned nothing"
    )

    console.print(f"[bold]rez python packages[/]:", highlight=False)
    for pythonVersion, pythonExecutable in rez_pip.rez.getPythonExecutables(
        args.python_version
    ).items():
        print(textwrap.indent(f"{pythonExecutable} ({pythonVersion})", prefix))

    print()

    rich.print(
        rich.panel.Panel(
            rich.text.Text(
                "Please redact any sensitive information before giving this output to someone else!\n\n"
                "Don't remove things, only redact/replace. Look for IP addresses, domain names, passwords, etc.",
                justify="center",
            ),
            title="[bold red]WARNING!",
            expand=False,
            border_style="yellow",
        ),
        file=sys.stderr,
    )


def run() -> int:
    pipWorkArea = tempfile.mkdtemp(prefix="rez-pip-target")
    args, pipArgs = _parseArgs(sys.argv[1:])

    try:
        _validateArgs(args)

        handler = rich.logging.RichHandler(
            show_time=False, markup=True, show_path=False
        )
        handler.setFormatter(logging.Formatter(fmt="%(message)s"))

        rootLogger = logging.getLogger("rez_pip")
        rootLogger.addHandler(handler)
        rootLogger.setLevel(args.log_level.upper())

        if args.debug_info:
            _debug(args)
            return 0

        _run(args, pipArgs, pipWorkArea)
        return 0
    except rez_pip.exceptions.RezPipError as exc:
        rich.get_console().print(exc, soft_wrap=True)
        return 1
    finally:
        if not args.keep_tmp_dirs:
            _LOG.debug(f"Removing {pipWorkArea}")
            shutil.rmtree(pipWorkArea)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "rez-pip"
copyright = "Contributors to the rez project"
author = "Rez Developers"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # first-party extensions
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    # Third-part
    "sphinx_inline_tabs",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]

# -- Options for sphinx.ext.autosectionlabel ---------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autosectionlabel.html

autosectionlabel_prefix_document = True



import os
import sys
import glob
import pathlib
import platform
import subprocess

import pytest
import installer.utils

import rez_pip.pip
import rez_pip.install

from . import utils


def test_installer_schemes():
    # Make sure we support all schemes and also future prrof in case installer adds
    # new schesm to its supported list of schemes.
    assert installer.utils.SCHEME_NAMES == (
        "purelib",
        "platlib",
        "headers",
        "scripts",
        "data",
    )


@pytest.mark.integration
def test_console_scripts(
    pythonRezPackage: str, rezRepo: str, tmp_path: pathlib.Path, index: utils.PyPIIndex
):
    executable, ctx = utils.getPythonRezPackageExecutablePath(pythonRezPackage, rezRepo)

    assert executable is not None

    installPath = tmp_path / "install"
    rez_pip.install.installWheel(
        rez_pip.pip.PackageInfo(
            rez_pip.pip.DownloadInfo(
                f"url",
                rez_pip.pip.ArchiveInfo(
                    "sha256=af266720050a66c893a6096a2f410989eeac74ff9a68ba194b3f6473e8e26171",
                    {
                        "sha256": "af266720050a66c893a6096a2f410989eeac74ff9a68ba194b3f6473e8e26171"
                    },
                ),
            ),
            False,
            True,
            rez_pip.pip.Metadata("0.1.0", "console_scripts"),
        ),
        index.getWheel("console_scripts"),
        os.fspath(installPath),
    )

    consoleScript = os.fspath(installPath / "scripts" / "console_scripts_cmd")
    if platform.system() == "Windows":
        consoleScript += ".exe"

    assert os.path.exists(consoleScript), f"{consoleScript!r} does not exists!"

    code, stdout, _ = ctx.execute_shell(
        command=[consoleScript],
        block=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        parent_environ={
            "PYTHONPATH": os.fspath(installPath / "python"),
            "SYSTEMROOT": os.environ.get("SYSTEMROOT", ""),
        },
        text=True,
    )

    # Use dirnames to avoid having to deal woth python vs python2 or python vs python3
    assert os.path.dirname(stdout) == os.path.dirname(
        executable
    ), f"stdout is {stdout!r} and executable is {executable!r}"



"""
Code that takes care of installing (extracting) wheels.
"""
import io
import os
import sys
import typing
import zipfile
import logging
import pathlib
import sysconfig

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

if typing.TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal

import installer
import installer.utils
import installer.records
import installer.scripts
import installer.sources
import installer.destinations

import rez_pip.pip

_LOG = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    LauncherKind = Literal["posix", "win-ia32", "win-amd64", "win-arm", "win-arm64"]
    ScriptSection = Literal["console", "gui"]


def isWheelPure(source: installer.sources.WheelSource) -> bool:
    stream = source.read_dist_info("WHEEL")
    metadata = installer.utils.parse_metadata_file(stream)
    return typing.cast(str, metadata["Root-Is-Purelib"]) == "true"


# Taken from https://github.com/pypa/installer/blob/main/src/installer/__main__.py#L49
def getSchemeDict(name: str, target: str) -> typing.Dict[str, str]:
    vars = {}
    vars["base"] = vars["platbase"] = installed_base = target

    schemeDict = sysconfig.get_paths(vars=vars)
    # calculate 'headers' path, not currently in sysconfig - see
    # https://bugs.python.org/issue44445. This is based on what distutils does.
    # TODO: figure out original vs normalised distribution names
    schemeDict["headers"] = os.path.join(
        sysconfig.get_path("include", vars={"installed_base": installed_base}),
        name,
    )

    if target:  # In practice this will always be set.
        schemeDict["purelib"] = os.path.join(target, "python")
        schemeDict["platlib"] = os.path.join(target, "python")
        schemeDict["headers"] = os.path.join(target, "headers", name)
        schemeDict["scripts"] = os.path.join(target, "scripts")
        # Potentiall handle data?
    return schemeDict


def installWheel(
    package: rez_pip.pip.PackageInfo,
    wheelPath: pathlib.Path,
    targetPath: str,
) -> typing.Tuple[importlib_metadata.Distribution, bool]:
    # TODO: Technically, target should be optional. We will always want to install in "pip install --target"
    #       mode. So right now it's a CLI option for debugging purposes.

    destination = CustomWheelDestination(
        getSchemeDict(package.name, targetPath),
        # TODO: Use Python from rez package, or simply use "/usr/bin/env python"?
        interpreter=sys.executable,
        script_kind=installer.utils.get_launcher_kind(),
    )

    isPure = True
    _LOG.debug(f"Installing {wheelPath} into {targetPath!r}")
    with installer.sources.WheelFile.open(wheelPath) as source:
        isPure = isWheelPure(source)

        installer.install(
            source=source,
            destination=destination,
            # Additional metadata that is generated by the installation tool.
            additional_metadata={
                "INSTALLER": f"rez-pip {importlib_metadata.version(__package__)}".encode(
                    "utf-8"
                ),
            },
        )

    targetPathPython = os.path.join(targetPath, "python")

    # That's kind of dirty, but using any other method returns inconsistent results.
    # For example, importlib.metadata.Distribution.discover(path=['/path']) sometimes
    # won't find the freshly intalled package, even if it exists and everything.
    path = os.path.join(
        targetPathPython,
        f"{package.name.replace('-', '_')}-{package.version}.dist-info",
    )
    dist = importlib_metadata.Distribution.at(path)

    if not dist.files:
        path = os.path.join(
            targetPathPython,
            # Some packages like sphinx will have have a sphinx.dist-info instead of Sphinx.dist-info.
            f"{package.name.replace('-', '_').lower()}-{package.version}.dist-info",
        )
        dist = importlib_metadata.Distribution.at(path)
        if not dist.files:
            raise RuntimeError(f"{path!r} does not exist!")

    return dist, isPure


# TODO: Document where this code comes from.
class CustomWheelDestination(installer.destinations.SchemeDictionaryDestination):
    # Exactly the same as SchemeDictionaryDestination, but uses our custom Script class.
    def write_script(
        self, name: str, module: str, attr: str, section: "ScriptSection"
    ) -> installer.records.RecordEntry:
        """Write a script to invoke an entrypoint.
        :param name: name of the script
        :param module: module path, to load the entry point from
        :param attr: final attribute access, for the entry point
        :param section: Denotes the "entry point section" where this was specified.
            Valid values are ``"gui"`` and ``"console"``.
        :type section: str
        - Generates a launcher using :any:`Script.generate`.
        - Writes to the "scripts" scheme.
        - Uses :py:meth:`SchemeDictionaryDestination.write_to_fs` for the
          filesystem interaction.
        """
        script = Script(name, module, attr, section)
        script_name, data = script.generate(self.interpreter, self.script_kind)

        with io.BytesIO(data) as stream:
            entry = self.write_to_fs(
                installer.utils.Scheme("scripts"),
                script_name,
                stream,
                is_executable=True,
            )

            path = self._path_with_destdir(
                installer.utils.Scheme("scripts"), script_name
            )
            mode = os.stat(path).st_mode
            mode |= (mode & 0o444) >> 2
            os.chmod(path, mode)

            return entry


_SCRIPT_TEMPLATE = """\
# -*- coding: utf-8 -*-
import re
import sys
from {module} import {import_name}
if __name__ == "__main__":
    sys.argv[0] = re.sub(r"(-script\\.pyw|\\.exe)?$", "", sys.argv[0])
    sys.exit({func_path}())
"""


# TODO: Document where this code comes from.
class Script(installer.scripts.Script):
    def generate(
        self, executable: str, kind: "LauncherKind"
    ) -> typing.Tuple[str, bytes]:
        """Generate a launcher for this script.
        :param executable: Path to the executable to invoke.
        :param kind: Which launcher template should be used.
            Valid values are ``"posix"``, ``"win-ia32"``, ``"win-amd64"`` and
            ``"win-arm"``.
        :type kind: str
        :raises InvalidScript: if no appropriate template is available.
        :return: The name and contents of the launcher file.
        """
        launcher = self._get_launcher_data(kind)
        # shebang = _build_shebang(executable, forlauncher=bool(launcher))
        # TODO: Should we instead just pass that value to WheelDestination?
        shebang = b"#!/usr/bin/env python"
        code = _SCRIPT_TEMPLATE.format(
            module=self.module,
            import_name=self.attr.split(".")[0],
            func_path=self.attr,
        ).encode("utf-8")

        if launcher is None:
            return (self.name, shebang + b"\n" + code)

        stream = io.BytesIO()
        with zipfile.ZipFile(stream, "w") as zf:
            zf.writestr("__main__.py", code)
        name = f"{self.name}.exe"
        data = launcher + shebang + b"\n" + stream.getvalue()
        return (name, data)



import os
import sys
import logging
import pathlib
import argparse
import subprocess
import unittest.mock

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

import pytest
import rich.console

import rez_pip.cli
import rez_pip.pip
import rez_pip.rez
import rez_pip.exceptions


def test_parseArgs_empty():
    args, pipArgs = rez_pip.cli._parseArgs([])
    assert vars(args) == {
        "constraint": None,
        "keep_tmp_dirs": False,
        "log_level": "info",
        "packages": [],
        "pip": rez_pip.pip.getBundledPip(),
        "prefix": None,
        "python_version": "3.7+",
        "release": False,
        "requirement": None,
        "debug_info": False,
    }

    assert pipArgs == []


@pytest.mark.parametrize("packages", [["package-a"], ["package-a", "package-b"]])
def test_parseArgs_packages(packages):
    args, pipArgs = rez_pip.cli._parseArgs(packages)
    assert vars(args) == {
        "constraint": None,
        "keep_tmp_dirs": False,
        "log_level": "info",
        "packages": packages,
        "pip": rez_pip.pip.getBundledPip(),
        "prefix": None,
        "python_version": "3.7+",
        "release": False,
        "requirement": None,
        "debug_info": False,
    }

    assert pipArgs == []


@pytest.mark.parametrize("files", [["-r=req1"], ["--requirement=req1", "-r=req2"]])
def test_parseArgs_no_package_with_requirements(files):
    args, pipArgs = rez_pip.cli._parseArgs(files)
    assert vars(args) == {
        "constraint": None,
        "keep_tmp_dirs": False,
        "log_level": "info",
        "packages": [],
        "pip": rez_pip.pip.getBundledPip(),
        "prefix": None,
        "python_version": "3.7+",
        "release": False,
        "requirement": [req.split("=")[-1] for req in files],
        "debug_info": False,
    }

    assert pipArgs == []


def test_parseArgs_constraints():
    args, pipArgs = rez_pip.cli._parseArgs(["-c", "asd", "-c", "adasdasd"])
    assert vars(args) == {
        "constraint": ["asd", "adasdasd"],
        "keep_tmp_dirs": False,
        "log_level": "info",
        "packages": [],
        "pip": rez_pip.pip.getBundledPip(),
        "prefix": None,
        "python_version": "3.7+",
        "release": False,
        "requirement": None,
        "debug_info": False,
    }

    assert pipArgs == []


def test_parseArgs_pipArgs():
    args, pipArgs = rez_pip.cli._parseArgs(
        ["-l", "info", "--", "adasdasd", "--requirement", "asd.txt"]
    )
    assert vars(args) == {
        "constraint": None,
        "keep_tmp_dirs": False,
        "log_level": "info",
        "packages": [],
        "pip": rez_pip.pip.getBundledPip(),
        "prefix": None,
        "python_version": "3.7+",
        "release": False,
        "requirement": None,
        "debug_info": False,
    }

    assert pipArgs == ["adasdasd", "--requirement", "asd.txt"]


def test_validateArgs_pip_good(tmp_path: pathlib.Path):
    path = pathlib.Path(tmp_path / "asd.pyz")

    path.touch()

    rez_pip.cli._validateArgs(
        argparse.Namespace(pip=os.fspath(path), packages=["asd"], requirement=None)
    )


def test_validateArgs_pip_bad(tmp_path: pathlib.Path):
    path = tmp_path / "asd.txt"

    path.touch()

    with pytest.raises(rez_pip.exceptions.RezPipError):
        rez_pip.cli._validateArgs(
            argparse.Namespace(pip=os.fspath(path), packages=["asd"], requirement=None)
        )


def test_validateArgs_pip_not_exists(tmp_path: pathlib.Path):
    path = tmp_path / "asd.pyz"

    with pytest.raises(rez_pip.exceptions.RezPipError) as exc:
        rez_pip.cli._validateArgs(
            argparse.Namespace(pip=os.fspath(path), packages=["asd"], requirement=None)
        )

    assert exc.value.message == f"zipapp at {os.fspath(path)!r} does not exist"


def test_validateArgs_no_packages_or_requirement(tmp_path: pathlib.Path):
    path = tmp_path / "asd.pyz"

    path.touch()

    with pytest.raises(
        rez_pip.exceptions.RezPipError,
        match="no packages were passed and --requirements was not used. At least one of them must be passed.",
    ) as exc:
        rez_pip.cli._validateArgs(
            argparse.Namespace(pip=os.fspath(path), packages=None, requirement=None)
        )


@pytest.fixture()
def resetLogger():
    logger = logging.getLogger("rez_pip")
    level = logger.level
    handlers = logger.handlers

    yield

    logger = logging.getLogger("rez_pip")
    logger.setLevel(level)

    for handler in logger.handlers:
        logger.removeHandler(handler)

    for handler in handlers:
        logger.addHandler(handler)


@pytest.mark.usefixtures("resetLogger")
def test_run_without_arguments(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(sys, "argv", ["rez-pip"])
    assert rez_pip.cli.run() == 1


@pytest.mark.usefixtures("resetLogger")
def test_run(monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory):
    monkeypatch.setattr(sys, "argv", ["rez-pip", "example"])

    tmppath = tmp_path_factory.mktemp("test_run")

    def mkdtemp(*args, **kwargs) -> str:
        return os.fspath(tmppath)

    with unittest.mock.patch("rez_pip.cli._run") as mocked:
        with unittest.mock.patch(
            "rez_pip.cli.tempfile.mkdtemp", side_effect=mkdtemp
        ) as mockedMkdtemp:
            assert rez_pip.cli.run() == 0

    assert mocked.called
    assert mockedMkdtemp.called

    assert not tmppath.exists()


@pytest.mark.usefixtures("resetLogger")
def test_run_removes_tmp_dirs_even_with_exceptions(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
):
    monkeypatch.setattr(sys, "argv", ["rez-pip", "example"])

    tmppath = tmp_path_factory.mktemp("test_run")

    def mkdtemp(*args, **kwargs) -> str:
        return os.fspath(tmppath)

    with unittest.mock.patch("rez_pip.cli._run", side_effect=RuntimeError) as mocked:
        with unittest.mock.patch(
            "rez_pip.cli.tempfile.mkdtemp", side_effect=mkdtemp
        ) as mockedMkdtemp:
            with pytest.raises(RuntimeError):
                assert rez_pip.cli.run() == 0

    assert mocked.called
    assert mockedMkdtemp.called

    assert not tmppath.exists()


@pytest.mark.usefixtures("resetLogger")
def test_run_keep_tmp_dirs_even_with_exceptions(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
):
    monkeypatch.setattr(sys, "argv", ["rez-pip", "example", "--keep-tmp-dirs"])

    tmppath = tmp_path_factory.mktemp("test_run")

    def mkdtemp(*args, **kwargs) -> str:
        return os.fspath(tmppath)

    with unittest.mock.patch("rez_pip.cli._run", side_effect=RuntimeError) as mocked:
        with unittest.mock.patch(
            "rez_pip.cli.tempfile.mkdtemp", side_effect=mkdtemp
        ) as mockedMkdtemp:
            with pytest.raises(RuntimeError):
                assert rez_pip.cli.run() == 0

    assert mocked.called
    assert mockedMkdtemp.called

    assert tmppath.exists()


@pytest.mark.usefixtures("resetLogger")
def test_run_keep_tmp_dirs(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
):
    monkeypatch.setattr(sys, "argv", ["rez-pip", "example", "--keep-tmp-dirs"])

    tmppath = tmp_path_factory.mktemp("test_run")

    def mkdtemp(*args, **kwargs) -> str:
        return os.fspath(tmppath)

    with unittest.mock.patch("rez_pip.cli._run") as mocked:
        with unittest.mock.patch(
            "rez_pip.cli.tempfile.mkdtemp", side_effect=mkdtemp
        ) as mockedMkdtemp:
            assert rez_pip.cli.run() == 0

    assert mocked.called
    assert mockedMkdtemp.called

    assert tmppath.exists()


@pytest.mark.usefixtures("resetLogger")
def test_run_with_debug_info(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(sys, "argv", ["rez-pip", "example", "--debug-info"])
    with unittest.mock.patch("rez_pip.cli._run") as mockedRun:
        with unittest.mock.patch("rez_pip.cli._debug") as mockedDebug:
            assert rez_pip.cli.run() == 0

    assert not mockedRun.called
    assert mockedDebug.called


@pytest.mark.skip("TODO: Figure out why it is failing")
def test_debug(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(
        rez_pip.rez,
        "getPythonExecutables",
        lambda x: {
            "3.7.15": "/path/to/python-3.7.15",
            "3.100.7": "/path/to/another/python-3.100.7",
        },
    )

    monkeypatch.setitem(os.environ, "TERM", "dumb")
    with unittest.mock.patch(
        "subprocess.run",
        side_effect=(
            unittest.mock.Mock(stdout="mocked pip --version"),
            unittest.mock.Mock(stdout="mocked pip config debug"),
            unittest.mock.Mock(stdout="mocked pip config list"),
        ),
    ) as mocked:
        rez_pip.cli._debug(
            argparse.Namespace(pip=None, python_version="2.7+"),
            console=rich.console.Console(),
        )

    assert mocked.call_args_list == [
        unittest.mock.call(
            [sys.executable, rez_pip.pip.getBundledPip(), "--version"],
            stdout=subprocess.PIPE,
            text=True,
        ),
        unittest.mock.call(
            [sys.executable, rez_pip.pip.getBundledPip(), "config", "debug"],
            stdout=subprocess.PIPE,
            text=True,
        ),
        unittest.mock.call(
            [sys.executable, rez_pip.pip.getBundledPip(), "config", "list"],
            stdout=subprocess.PIPE,
            text=True,
        ),
    ]

    captured = capsys.readouterr()

    assert (
        captured.out
        == f"""rez-pip version: {importlib_metadata.version("rez-pip")}
rez version: 2.112.0
python version: {sys.version}
python executable: {sys.executable}
pip: {rez_pip.pip.getBundledPip()}
pip version: mocked pip --version
rez-pip provided arguments:
  {{
      "pip": null,
      "python_version": "2.7+"
  }}
pip config debug:
  mocked pip config debug
pip config list:
  mocked pip config list
rez python packages:
  /path/to/python-3.7.15 (3.7.15)
  /path/to/another/python-3.100.7 (3.100.7)

"""
    )



import os
import re
import sys
import uuid
import typing
import pathlib
import subprocess

import rich
import pytest

import rez_pip.pip
import rez_pip.exceptions

from . import utils


def test_getBundledPip():
    """Test that the bundled pip exists and can be executed"""
    assert os.path.exists(rez_pip.pip.getBundledPip())

    subprocess.run([sys.executable, rez_pip.pip.getBundledPip(), "-h"])


@pytest.mark.parametrize(
    "packages,expectedPackages",
    [
        [
            ["package_a"],
            [
                rez_pip.pip.PackageInfo(
                    rez_pip.pip.DownloadInfo(
                        "{pypi}/packages/package_a/package_a-1.0.0-py2.py3-none-any.whl",
                        rez_pip.pip.ArchiveInfo(
                            "sha256=<val>",
                            {"sha256": "<val>"},
                        ),
                    ),
                    False,
                    True,
                    rez_pip.pip.Metadata("1.0.0", "package_a"),
                ),
            ],
        ],
        [
            ["package_a", "console_scripts"],
            [
                rez_pip.pip.PackageInfo(
                    rez_pip.pip.DownloadInfo(
                        "{pypi}/packages/package_a/package_a-1.0.0-py2.py3-none-any.whl",
                        rez_pip.pip.ArchiveInfo(
                            "sha256=<val>",
                            {"sha256": "<val>"},
                        ),
                    ),
                    False,
                    True,
                    rez_pip.pip.Metadata("1.0.0", "package_a"),
                ),
                rez_pip.pip.PackageInfo(
                    download_info=rez_pip.pip.DownloadInfo(
                        url="{pypi}/packages/console_scripts/console_scripts-0.1.0-py2.py3-none-any.whl",
                        archive_info=rez_pip.pip.ArchiveInfo(
                            hash="sha256=<val>",
                            hashes={"sha256": "<val>"},
                        ),
                    ),
                    is_direct=False,
                    requested=True,
                    metadata=rez_pip.pip.Metadata(
                        version="0.1.0", name="console_scripts"
                    ),
                ),
            ],
        ],
    ],
    ids=["package_a", "package_a+console_scripts"],
)
def test_getPackages_no_deps(
    packages: typing.List[str],
    expectedPackages: typing.List[rez_pip.pip.PackageInfo],
    pythonRezPackage: str,
    rezRepo: str,
    pypi: str,
    index: utils.PyPIIndex,
):
    """
    This just tests that the function returns PackageInfo objects
    and it's why we use --no-deps. The scenario with dependencies
    will be tested in another test.
    """
    executable, ctx = utils.getPythonRezPackageExecutablePath(pythonRezPackage, rezRepo)
    assert executable is not None

    for expectedPackage in expectedPackages:
        expectedPackage.download_info.archive_info.hash = (
            f"sha256={index.getWheelHash(expectedPackage.name)}"
        )
        expectedPackage.download_info.archive_info.hashes = {
            "sha256": index.getWheelHash(expectedPackage.name)
        }

    resolvedPackages = rez_pip.pip.getPackages(
        packages,
        rez_pip.pip.getBundledPip(),
        "3.11",
        executable,
        [],
        [],
        ["--index-url", pypi, "-vvv", "--no-deps", "--retries=0"],
    )

    for packageInfo in expectedPackages:
        packageInfo.download_info.url = packageInfo.download_info.url.format(pypi=pypi)

    assert resolvedPackages == expectedPackages


def test_getPackages_with_deps(
    pythonRezPackage: str,
    rezRepo: str,
    pypi: str,
):
    """
    Test that we get all dependencies
    """
    executable, ctx = utils.getPythonRezPackageExecutablePath(pythonRezPackage, rezRepo)
    assert executable is not None

    resolvedPackages = rez_pip.pip.getPackages(
        ["package_a"],
        rez_pip.pip.getBundledPip(),
        "3.11",
        executable,
        [],
        [],
        ["--index-url", pypi, "-vvv", "--retries=0"],
    )

    resolvedPackageNames = [pkg.name for pkg in resolvedPackages]

    assert sorted(resolvedPackageNames) == ["package_a", "package_b"]


def test_getPackages_error(
    pythonRezPackage: str, rezRepo: str, pypi: str, tmp_path: pathlib.Path
):
    executable, ctx = utils.getPythonRezPackageExecutablePath(pythonRezPackage, rezRepo)
    assert executable is not None

    packageName = str(uuid.uuid4())

    with pytest.raises(rez_pip.exceptions.PipError) as exc:
        rez_pip.pip.getPackages(
            [packageName],
            rez_pip.pip.getBundledPip(),
            ".".join(str(i) for i in sys.version_info[:2]),
            executable,
            [],
            [],
            # Disable retries to speed up the test
            [
                # Specify index to avoid pip complaining about OpenSSL not being available.
                f"--index-url={pypi}",
                "--find-links",
                os.fspath(tmp_path),
                "-v",
                "--retries",
                "0",
            ],
        )

    with rich.get_console().capture() as capture:
        rich.get_console().print(exc.value, soft_wrap=True)

    match = re.match(
        r"rez_pip\.exceptions\.PipError: Failed to run pip command\: '.*'",
        capture.get().splitlines()[0],
    )

    assert match is not None

    # Lowercase to avoid discrepencies between C:\ and c:\
    assert (
        "\n".join(capture.get().splitlines()[1:]).lower()
        == f"""
Pip reported this:

Looking in indexes: {pypi}
Looking in links: {os.fspath(tmp_path)}
ERROR: Could not find a version that satisfies the requirement {packageName} (from versions: none)
ERROR: No matching distribution found for {packageName}""".lower()
    )


def test__readPipReport(tmp_path: pathlib.Path):
    # check for unicode encoding errors
    reportSrcContent = '{\n"description": "'
    reportSrcContent += "[English readme]"
    reportSrcContent += "•[简体中文 readme]"
    reportSrcContent += "•[正體中文 readme]"
    reportSrcContent += "•[Lengua española readme]"
    reportSrcContent += "•[Läs på svenska]"
    reportSrcContent += "•[日本語 readme]"
    reportSrcContent += "•[한국어 readme]"
    reportSrcContent += "•[Français readme]"
    reportSrcContent += "•[Schwizerdütsch readme"
    reportSrcContent += "•[हिन्दी readme]"
    reportSrcContent += "•[Русский readme]"
    reportSrcContent += "•[فارسی readme]•"
    reportSrcContent += "•[Türkçe readme]"
    reportSrcContent += '"\n}'

    reportPath = tmp_path / "report"
    reportPath.write_text(reportSrcContent, encoding="utf-8")

    reportContent = rez_pip.pip._readPipReport(reportPath=str(reportPath))
    assert reportContent



import os
import sys
import asyncio
import tempfile
import subprocess

import aiohttp
import gidgethub.aiohttp
import packaging.version

BUNDLED_PIP_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "src", "rez_pip", "data", "pip.pyz")
)
TMP_DIR = os.environ.get("RUNNER_TEMP", tempfile.gettempdir())


async def getLatestPip(session: aiohttp.ClientSession) -> packaging.version.Version:
    response = await session.get(
        "https://pypi.org/pypi/pip/json",
        headers={"Content-Type": "application/json"},
    )

    pypiInfo = await response.json()
    releases = pypiInfo["releases"]

    latest = sorted(releases, key=lambda x: packaging.version.parse(x))[-1]
    return packaging.version.parse(latest)


def getVersionFromZipApp(path: str) -> packaging.version.Version:
    bundledVersion = (
        subprocess.check_output(
            [
                "python",
                path,
                "--version",
            ],
            text=True,
        )
        .strip()
        .split(" ")[1]
    )

    return packaging.version.parse(bundledVersion)


async def createPullRequest(session: aiohttp.ClientSession) -> None:
    pass


async def updatePullRequest(session: aiohttp.ClientSession, pr: dict) -> None:
    pass


async def downloadNewPip(
    session: aiohttp.ClientSession, version: packaging.version.Version
) -> str:
    downloadPath = os.path.join(TMP_DIR, f"pip-{version}.pyz")

    async with session.get(
        # TODO: Use specific version once https://github.com/pypa/get-pip/issues/189 gets merged and released.
        "https://bootstrap.pypa.io/pip/pip.pyz",
        headers={
            "Content-Type": "application/octet-stream",
            "User-Agent": "rez-pip/0.1.0",
        },
    ) as response:
        with open(downloadPath, "wb") as fd:
            fd.write(await response.read())

    downloadedVersion = getVersionFromZipApp(downloadPath)

    assert (
        downloadedVersion == version
    ), f"The downloaded pip version is {downloadedVersion} but expected it to be {version}"

    return downloadPath


def createGHAOutputVariable(name: str, value: str):
    if githubOutput := os.environ.get("GITHUB_OUTPUT", ""):
        with open(githubOutput, "a", encoding="utf-8") as fd:
            fd.write(f"{name}={value}\n")


async def main():
    async with aiohttp.ClientSession() as session:
        latestPipVersion = await getLatestPip(session)
        print(f"latest pip version on PyPI is {latestPipVersion}")

        bundledPipVersion = getVersionFromZipApp(BUNDLED_PIP_PATH)
        print(f"The bundled pip version is {bundledPipVersion}")

        createGHAOutputVariable("previous-pip-version", str(bundledPipVersion))

        if latestPipVersion <= bundledPipVersion:
            createGHAOutputVariable("downloaded-pip-path", "none")
            createGHAOutputVariable("new-pip-version", "none")
            return

        downloadPath = await downloadNewPip(session, latestPipVersion)
        print(
            f"Downloaded pip {latestPipVersion} to {downloadPath}. The path will be accessible as a step output using downloaded-pip-path."
        )

        createGHAOutputVariable("downloaded-pip-path", downloadPath)
        createGHAOutputVariable("new-pip-version", str(latestPipVersion))


if __name__ == "__main__":
    sys.exit(bool(asyncio.run(main())))



"""Development automation"""
import nox

# nox.options.sessions = ["lint", "test", "doctest"]
nox.options.reuse_existing_virtualenvs = True


@nox.session()
def lint(session: nox.Session):
    pass


@nox.session()
def mypy(session: nox.Session):
    session.install("mypy")
    session.install(".", "-c", "tests/constraints.txt")

    session.run("mypy")


@nox.session()
def format(session: nox.Session):
    session.install("black")

    session.run("black", ".", "--check")


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def test(session: nox.Session):
    session.install("-r", "tests/requirements.txt")
    session.install(".")

    session.run(
        "pytest",
        *session.posargs,
    )


@nox.session()
def download_pip(session: nox.Session):
    session.install("packaging", "gidgethub", "aiohttp")

    session.run("python", "./scripts/download_pip.py")



import os
import sys
import json
import typing
import logging
import tempfile
import itertools
import subprocess
import dataclasses

import dataclasses_json

import rez_pip.data
import rez_pip.exceptions

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Metadata(dataclasses_json.DataClassJsonMixin):
    version: str
    name: str


@dataclasses.dataclass
class ArchiveInfo(dataclasses_json.DataClassJsonMixin):
    hash: str
    hashes: typing.Dict[str, str]


@dataclasses.dataclass
class DownloadInfo(dataclasses_json.DataClassJsonMixin):
    url: str
    archive_info: ArchiveInfo

    dataclass_json_config = dataclasses_json.config(
        undefined=dataclasses_json.Undefined.EXCLUDE
    )


@dataclasses.dataclass
class PackageInfo(dataclasses_json.DataClassJsonMixin):
    download_info: DownloadInfo
    is_direct: bool
    requested: bool
    metadata: Metadata

    dataclass_json_config = dataclasses_json.config(
        undefined=dataclasses_json.Undefined.EXCLUDE
    )

    @property
    def name(self) -> str:
        return self.metadata.name

    @property
    def version(self) -> str:
        return self.metadata.version


def getBundledPip() -> str:
    return os.path.join(os.path.dirname(rez_pip.data.__file__), "pip.pyz")


def getPackages(
    packageNames: typing.List[str],
    pip: str,
    pythonVersion: str,
    pythonExecutable: str,
    requirements: typing.List[str],
    constraints: typing.List[str],
    extraArgs: typing.List[str],
) -> typing.List[PackageInfo]:
    # python pip.pyz install -q requests --dry-run --ignore-installed --python-version 2.7 --only-binary=:all: --target /tmp/asd --report -

    _fd, tmpFile = tempfile.mkstemp(prefix="pip-install-output", text=True)
    os.close(_fd)
    # We can't with "with" (context manager) because it will fail on Windows.
    # Windows doesn't allow two different processes to write if the file is
    # already opened.
    try:
        command = [
            # We need to use the real interpreter because pip can't resolve
            # markers correctly even if --python-version is provided.
            # See https://github.com/pypa/pip/issues/11664.
            pythonExecutable,
            pip,
            "install",
            "-q",
            *packageNames,
            *list(itertools.chain(*zip(["-r"] * len(requirements), requirements))),
            *list(itertools.chain(*zip(["-c"] * len(constraints), constraints))),
            "--disable-pip-version-check",
            "--dry-run",
            "--ignore-installed",
            f"--python-version={pythonVersion}" if pythonVersion else "",
            "--only-binary=:all:",
            "--target=/tmp/asd",
            "--disable-pip-version-check",
            "--report",  # This is the "magic". Pip will generate a JSON with all the resolved URLs.
            tmpFile,
            *extraArgs,
        ]

        _LOG.debug(f"Running {' '.join(command)!r}")
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

        pipOutput = []
        while True:
            stdout = typing.cast(typing.IO[str], process.stdout).readline()
            if process.poll() is not None:
                break
            if stdout:
                pipOutput.append(stdout.rstrip())
                sys.stdout.write(stdout)

        if process.poll() != 0:
            output = "\n".join(pipOutput)
            raise rez_pip.exceptions.PipError(
                f"[bold red]Failed to run pip command[/]: {' '.join(command)!r}\n\n"
                "[bold]Pip reported this[/]:\n\n"
                f"{output}",
            )
        reportContent = _readPipReport(reportPath=tmpFile)
    finally:
        os.remove(tmpFile)

    rawPackages = reportContent["install"]

    packages: typing.List[PackageInfo] = []

    for rawPackage in rawPackages:
        packageInfo = PackageInfo.from_dict(rawPackage)
        packages.append(packageInfo)

    return packages


def _readPipReport(reportPath: str) -> typing.Dict[str, typing.Any]:
    """
    Retrieve the json report generated by pip as json dict object.
    """
    with open(reportPath, "r", encoding="utf-8") as reportFile:
        reportContent: typing.Dict[typing.Any, typing.Any] = json.load(reportFile)
    return reportContent
